# -*- coding: utf-8 -*-
"""
implicit version of dependency cache from wbia/templates/template_generator
"""
import logging
import os.path

import utool as ut
import numpy as np
import six
from six.moves import zip
from wbia.dtool import sql_control
from wbia.dtool import depcache_table
from wbia.dtool import base
from collections import defaultdict

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')


# global function registry
PREPROC_REGISTER = defaultdict(list)
SUBPROP_REGISTER = defaultdict(list)


REG_PREPROC_DOC = """
Args:
    tablename (str): name of the node (corrsponds to SQL table)
    parents (list): tables this node depends on
    colnames (list): data returned by this table
    coltypes (list): types of data returned by this table
    chunksize (int): (default = None)
    configclass (dtool.TableConfig): derivative of dtool.TableConfig.
        if None, a default class will be constructed for you. (default = None)
    docstr (str): (default = None)
    fname (str):  file name(default = None)
    asobject (bool): hacky dont use (default = False)

SeeAlso:
    depcache_table.DependencyCacheTable
"""


def check_register(args, kwargs):
    assert len(args) < 6, 'too many args'
    assert 'preproc_func' not in kwargs, 'cannot specify func in wrapper'


def make_depcache_decors(root_tablename):
    """
    Makes global decorators to register functions for a tablename.

    A preproc function is meant to belong only to a single parent An algo
    function belongs to the root node, and may depend on a set of root nodes
    rather than just a single one.
    """

    @ut.apply_docstr(REG_PREPROC_DOC)
    def register_preproc(*args, **kwargs):
        """
        Global regsiter proproc function that will define a table for all
        dependency caches containing the parents. See

        See dtool.depcache_control.REG_PREPROC_DOC if docstr is not autogened
        """

        def register_preproc_wrapper(func):
            check_register(args, kwargs)
            kwargs['preproc_func'] = func
            PREPROC_REGISTER[root_tablename].append((args, kwargs))
            return func

        return register_preproc_wrapper

    def register_subprop(*args, **kwargs):
        def _wrapper(func):
            kwargs['preproc_func'] = func
            SUBPROP_REGISTER[root_tablename].append((args, kwargs))
            return func

        return _wrapper

    _depcdecors = ut.odict({'preproc': register_preproc, 'subprop': register_subprop})
    return _depcdecors


class _CoreDependencyCache(object):
    """
    Core worker functions for the depcache
    Inherited by a calss with some "nice extras
    """

    @profile
    def _register_prop(
        depc,
        tablename,
        parents=None,
        colnames=None,
        coltypes=None,
        preproc_func=None,
        fname=None,
        configclass=None,
        requestclass=None,
        default_to_unpack=None,
        **kwargs,
    ):
        """
        Registers a table with this dependency cache.
        Essentially passes args down to make a DependencyTable.

        SEE: dtool.REG_PREPROC_DOC
        """
        if depc._debug:
            logger.info('[depc] Registering tablename=%r' % (tablename,))
            logger.info('[depc]  * preproc_func=%r' % (preproc_func,))
        # ----------
        # Sanitize inputs
        if isinstance(tablename, six.string_types):
            tablename = six.text_type(tablename)
        if parents is None:
            parents = [depc.root]
        if colnames is None:
            colnames = 'data'
            if coltypes is None:
                coltypes = np.ndarray
        # Check if just a single column is given
        if default_to_unpack is None:
            if ut.isiterable(colnames):
                default_to_unpack = False
                colnames = ut.lmap(six.text_type, colnames)
            else:
                colnames = [colnames]
                coltypes = [coltypes]
                default_to_unpack = True
        if coltypes is None:
            raise ValueError('must specify coltypes of %s' % (tablename,))
            coltypes = [np.ndarray] * len(colnames)
        if fname is None:
            fname = depc.default_fname
        if configclass is None:
            # Make a default config with no parameters
            configclass = {}
        if isinstance(configclass, dict):
            # Dynamically make config class
            default_cfgdict = configclass
            configclass = base.make_configclass(default_cfgdict, tablename)
        # ----------
        # Register a new table and configuration
        if requestclass is not None:
            depc.requestclass_dict[tablename] = requestclass
        depc.fname_to_db[fname] = None
        table = depcache_table.DependencyCacheTable(
            depc=depc,
            parent_tablenames=parents,
            tablename=tablename,
            data_colnames=colnames,
            data_coltypes=coltypes,
            preproc_func=preproc_func,
            fname=fname,
            default_to_unpack=default_to_unpack,
            **kwargs,
        )
        depc.cachetable_dict[tablename] = table
        depc.configclass_dict[tablename] = configclass
        return table

    @ut.apply_docstr(REG_PREPROC_DOC)
    def register_preproc(depc, *args, **kwargs):
        """
        Decorator for registration of cachables
        """

        def register_preproc_wrapper(func):
            check_register(args, kwargs)
            kwargs['preproc_func'] = func
            depc._register_prop(*args, **kwargs)
            return func

        return register_preproc_wrapper

    def _register_subprop(depc, tablename, propname=None, preproc_func=None):
        """ subproperties are always recomputeed on the fly """
        table = depc.cachetable_dict[tablename]
        table.subproperties[propname] = preproc_func

    def close(depc):
        """
        Close all managed SQL databases
        """
        for fname, db in depc.fname_to_db.items():
            db.close()

    @profile
    def initialize(depc, _debug=None):
        """
        Creates all registered tables
        """
        logger.info(
            '[depc] Initialize %s depcache in %r' % (depc.root.upper(), depc.cache_dpath)
        )
        _debug = depc._debug if _debug is None else _debug
        if depc._use_globals:
            reg_preproc = PREPROC_REGISTER[depc.root]
            reg_subprop = SUBPROP_REGISTER[depc.root]
            if ut.VERBOSE:
                logger.info(
                    '[depc.init] Registering %d global preproc funcs' % len(reg_preproc)
                )
            for args_, _kwargs in reg_preproc:
                depc._register_prop(*args_, **_kwargs)
            if ut.VERBOSE:
                logger.info(
                    '[depc.init] Registering %d global subprops ' % len(reg_subprop)
                )
            for args_, _kwargs in reg_subprop:
                depc._register_subprop(*args_, **_kwargs)

        ut.ensuredir(depc.cache_dpath)

        # Memory filestore
        # if False:
        #    # http://docs.pyfilesystem.org/en/latest/getting_started.html
        #    pip install fs

        for fname in depc.fname_to_db.keys():
            if fname == ':memory:':
                fpath = fname
            else:
                fname_ = ut.ensure_ext(fname, '.sqlite')
                from os.path import dirname

                prefix_dpath = dirname(fname_)
                if prefix_dpath:
                    ut.ensuredir(ut.unixjoin(depc.cache_dpath, prefix_dpath))
                fpath = ut.unixjoin(depc.cache_dpath, fname_)
            # if ut.get_argflag('--clear-all-depcache'):
            #     ut.delete(fpath)
            db_uri = 'file://{}'.format(os.path.realpath(fpath))
            db = sql_control.SQLDatabaseController.from_uri(db_uri)
            depcache_table.ensure_config_table(db)
            depc.fname_to_db[fname] = db
        if ut.VERBOSE:
            logger.info('[depc] Finished initialization')

        for table in depc.cachetable_dict.values():
            table.initialize(_debug=_debug)

        # HACKS:
        # Define injected functions for autocomplete convinience
        class InjectedDepc(object):
            pass

        depc.d = InjectedDepc()
        depc.w = InjectedDepc()
        d = depc.d
        w = depc.w
        inject_patterns = [
            ('get_{tablename}_rowids', depc.get_rowids),
            ('get_{tablename}_config_history', depc.get_config_history),
        ]
        for table in depc.cachetable_dict.values():
            wobj = InjectedDepc()
            # Set nested version
            setattr(w, table.tablename, wobj)
            for dfmtstr, func in inject_patterns:
                funcname = ut.get_funcname(func)
                attrname = dfmtstr.format(tablename=table.tablename)
                get_rowids = ut.partial(func, table.tablename)
                # Set flat version
                setattr(d, attrname, get_rowids)
                setattr(wobj, funcname, func)
            dfmtstr = 'get_{tablename}_{colname}'
            for colname in table.data_colnames:
                get_prop = ut.partial(depc.get, table.tablename, colnames=colname)
                attrname = dfmtstr.format(tablename=table.tablename, colname=colname)
                # Set flat version
                setattr(d, attrname, get_prop)
                setattr(wobj, 'get_' + colname, get_prop)

    # -----------------------------
    # GRAPH INSPECTION

    def get_dependencies(depc, tablename):
        """
        gets level dependences from root to tablename

        CommandLine:
            python -m dtool.depcache_control --exec-get_dependencies

        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.dtool.depcache_control import *  # NOQA
            >>> from wbia.dtool.example_depcache import testdata_depc
            >>> depc = testdata_depc()
            >>> tablename = 'fgweight'
            >>> result = ut.repr3(depc.get_dependencies(tablename), nl=1)
            >>> print(result)
            [
                ['dummy_annot'],
                ['chip', 'probchip'],
                ['keypoint'],
                ['fgweight'],
            ]
        """
        try:
            assert tablename in depc.cachetable_dict, 'tablename=%r does not exist' % (
                tablename,
            )
            root = depc.root_tablename
            children_, parents_ = list(zip(*depc.get_edges()))
            child_to_parents = ut.group_items(children_, parents_)
            if ut.VERYVERBOSE:
                logger.info('root = %r' % (root,))
                logger.info('tablename = %r' % (tablename,))
                logger.info('child_to_parents = %s' % (ut.repr3(child_to_parents),))
            to_root = {tablename: ut.paths_to_root(tablename, root, child_to_parents)}
            if ut.VERYVERBOSE:
                logger.info('to_root = %r' % (to_root,))
            from_root = ut.reverse_path(to_root, root, child_to_parents)
            dependency_levels_ = ut.get_levels(from_root)
            dependency_levels = ut.longest_levels(dependency_levels_)
            dependency_levels = list(map(sorted, dependency_levels))
        except Exception as ex:
            ut.printex(
                ex,
                'error getting dependencies',
                keys=[
                    'tablename',
                    'root',
                    'children_to_parents',
                    'to_root',
                    'from_root',
                    'dependency_levels_',
                    'dependency_levels',
                ],
            )
            raise

        return dependency_levels

    def _ensure_config(depc, tablekey, config, _debug=False):
        """
        Creates a full table configuration with all defaults using config

        Args:
            tablekey (str): name of the table to grab config from
            config (dict): may be overspecified or underspecfied
        """
        configclass = depc.configclass_dict.get(tablekey, None)
        # requestclass = depc.requestclass_dict.get(tablekey, None)
        if configclass is None:
            config_ = config
        else:
            # Grab the correct configclass for the current table
            config_ = None
            if config is None:
                config_ = configclass()
            elif len(getattr(config, '_subconfig_attrs', [])) > 0:
                # Get correct config for implicit dependencies
                target_name = configclass().get_config_name()
                if target_name in config._subconfig_names:
                    _index = config._subconfig_names.index(target_name)
                    subcfg_attr = config._subconfig_attrs[_index]
                    config_ = config[subcfg_attr]
            if config_ is None:
                # Preferable way to get configs with explicit
                # configs
                if _debug:
                    logger.info(' **config = %r' % (config,))
                config_ = configclass(**config)
                if _debug:
                    logger.info(' config_ = %r' % (config_,))
        return config_

    def get_config_trail(depc, tablename, config):
        graph = depc.make_graph(implicit=True)
        tablename_list = ut.nx_all_nodes_between(graph, depc.root, tablename)
        tablename_list = ut.nx_topsort_nodes(graph, tablename_list)
        config_trail = []
        for tablekey in tablename_list:
            if tablekey in depc.configclass_dict:
                config_ = depc._ensure_config(tablekey, config)
                config_trail.append(config_)
        return config_trail

    def get_config_trail_str(depc, tablename, config):
        config_trail = depc.get_config_trail(tablename, config)
        trail_cfgstr = '_'.join([x.get_cfgstr() for x in config_trail])
        return trail_cfgstr

    def _get_parent_input(
        depc,
        tablename,
        root_rowids,
        config,
        ensure=True,
        _debug=None,
        recompute=False,
        recompute_all=False,
        eager=True,
        nInput=None,
    ):
        # Get ancestor rowids that are descendants of root
        table = depc[tablename]
        rowid_dict = depc.get_all_descendant_rowids(
            tablename,
            root_rowids,
            config=config,
            ensure=ensure,
            eager=eager,
            nInput=nInput,
            recompute=recompute,
            recompute_all=recompute_all,
            _debug=ut.countdown_flag(_debug),
            levels_up=1,
        )
        parent_rowids = depc._get_parent_rowids(table, rowid_dict)
        return parent_rowids

    # -----------------------------
    # STATE GETTERS

    def rectify_input_tuple(depc, exi_inputs, input_tuple):
        """
        Standardizes inputs allowed for convinience into the expected input for
        get_parent_rowids.
        """
        input_tuple_ = input_tuple

        if isinstance(input_tuple_, (list, np.ndarray)):
            if len(exi_inputs) != 1:
                msg = '#expected=%d, #got=1' % (len(exi_inputs),)
                msg += '. Did you forget to cast multi-inputs to a tuple?'
                raise ValueError(msg)
            # Inputs should always be a tuple of lists.
            input_tuple_ = (input_tuple_,)
        if len(exi_inputs) == 1:
            # HACK: for simple case where we only need one parent
            if isinstance(input_tuple_, (tuple,)):
                if len(input_tuple_) == 0:
                    input_tuple_ = []
                elif len(input_tuple_) > 1:
                    if not ut.isiterable(input_tuple_[0]):
                        input_tuple_ = (input_tuple_,)
        if len(exi_inputs) != len(input_tuple_):
            msg = '#expected=%d, #got=%d' % (len(exi_inputs), len(input_tuple_))
            raise ValueError(msg)

        # rectify input depth
        rectified_input = []
        for x, d in zip(input_tuple_, exi_inputs.expected_input_depth()):
            if d == 0:
                if not ut.isiterable(x):
                    rectified_input.append([x])
                else:
                    rectified_input.append(x)
            else:
                if ut.list_depth(x) == 0:
                    rectified_input.append([x])
                else:
                    rectified_input.append(x)
        return rectified_input

    def get_parent_rowids(depc, target_tablename, input_tuple, config=None, **kwargs):
        """
        Returns the parent rowids needed to get / compute a property of
        tablename

        Args:
            input_tuple :
                to be explicit send in as a tuple of lists.  Each list
                corresponds to parent information needed by expanded rmis (root
                most input).

                Each item in the tuple correponds a root most node, and should
                be specified as a list of inputs. For single items this is a
                scalar, for multi-items it is a list.

                For example if you have a property like a chip that depends on
                only one parent, then to get the chips for the first N
                annotations your list input tuple is:
                    input_tuple = ([1, 2, 3, ..., N],)

                For a single multi inputs:
                If you want to get two vocabs for even and odd annots then you
                have:
                    ([[0, 2, 4, ...], [1, 3, 5, ...]],)

                For a single comparasion version multi inputs:
                If you want to query the first N annotats against two
                vocabs then you have:
                    ([1, 2, 3, ..., N], [[0, 2, 4, ...], [1, 3, 5, ...]],)
                (Note this only works if broadcasting is on)
        """
        _kwargs = kwargs.copy()
        _recompute = _kwargs.pop('recompute_all', False)
        _debug = _kwargs.get('_debug', False)
        _hack_rootmost = _kwargs.pop('_hack_rootmost', False)
        _debug = depc._debug if _debug is None else _debug
        if config is None:
            config = {}

        with ut.Indenter('[GetParentID-%s]' % (target_tablename,), enabled=_debug):
            if _debug:
                logger.info(ut.color_text('Enter get_parent_rowids', 'blue'))
                logger.info(' * target_tablename = %r' % (target_tablename,))
                logger.info(' * input_tuple=%s' % (ut.trunc_repr(input_tuple),))
                logger.info(' * config = %r' % (config,))
            target_table = depc[target_tablename]

            # TODO: Expand to the appropriate given inputs
            if _hack_rootmost:
                # Hack: if true, we are given inputs in rootmost form
                exi_inputs = target_table.rootmost_inputs
            else:
                # otherwise we are given inputs in totalroot form
                exi_inputs = target_table.rootmost_inputs.total_expand()
            if _debug:
                logger.info(' * exi_inputs=%s' % (exi_inputs,))

            rectified_input = depc.rectify_input_tuple(exi_inputs, input_tuple)

            rowid_dict = {}
            for rmi, rowids in zip(exi_inputs.rmi_list, rectified_input):
                rowid_dict[rmi] = rowids

            compute_edges = exi_inputs.flat_compute_rmi_edges()
            if _debug:
                logger.info(' * rectified_input=%s' % ut.trunc_repr(rectified_input))
                logger.info(' * compute_edges=%s' % ut.repr2(compute_edges, nl=2))

            for count, (input_nodes, output_node) in enumerate(compute_edges, start=1):
                if _debug:
                    ut.cprint(
                        ' * COMPUTING %d/%d EDGE %r -- %r'
                        % (count, len(compute_edges), input_nodes, output_node),
                        'blue',
                    )
                tablekey = output_node.tablename
                table = depc[tablekey]
                input_nodes_ = input_nodes
                if _debug:
                    logger.info(
                        'table.parent_id_tablenames = %r' % (table.parent_id_tablenames,)
                    )
                    logger.info('input_nodes_ = %r' % (input_nodes_,))
                input_multi_flags = [
                    node.ismulti and node in exi_inputs.rmi_list for node in input_nodes_
                ]

                # Args currently go in like this:
                # args  = [..., (pid_{i,1}, pid_{i,2}, ..., pid_{i,M}), ...]
                # They get converted into
                # argsT = [... (pid_{1,j}, ... pid_{N,j}) ...]
                # i = row, j = col
                sig_multi_flags = table.get_parent_col_attr('ismulti')
                parent_rowidsT = ut.take(rowid_dict, input_nodes_)
                parent_rowids_ = []
                # TODO: will need to figure out which columns to zip and which
                # columns to product (ie take product over ones that have 1
                # item, and zip ones that have equal amount of items)
                for flag1, flag2, rowidsT in zip(
                    sig_multi_flags, input_multi_flags, parent_rowidsT
                ):
                    if flag1 and flag2:
                        parent_rowids_.append(rowidsT)
                    elif flag1 and not flag2:
                        parent_rowids_.append([rowidsT])
                    elif not flag1 and flag2:
                        assert len(rowidsT) == 1
                        parent_rowids_.append(rowidsT[0])
                    else:
                        parent_rowids_.append(rowidsT)
                # Assume that we are either given corresponding lists or single values
                # that must be broadcast.
                rowlens = list(map(len, parent_rowids_))
                maxlen = max(rowlens)
                parent_rowids2_ = [
                    r * maxlen if len(r) == 1 else r for r in parent_rowids_
                ]
                _parent_rowids = list(zip(*parent_rowids2_))
                # _parent_rowids = list(ut.product(*parent_rowids_))

                if _debug:
                    logger.info(
                        'parent_rowids_ = %s'
                        % (
                            ut.repr4(
                                [ut.trunc_repr(ids_) for ids_ in parent_rowids_],
                                strvals=True,
                            )
                        )
                    )
                    logger.info(
                        'parent_rowids2_ = %s'
                        % (
                            ut.repr4(
                                [ut.trunc_repr(ids_) for ids_ in parent_rowids2_],
                                strvals=True,
                            )
                        )
                    )
                    logger.info(
                        '_parent_rowids = %s'
                        % (
                            ut.truncate_str(
                                ut.repr4(
                                    [ut.trunc_repr(ids_) for ids_ in _parent_rowids],
                                    strvals=True,
                                )
                            )
                        )
                    )

                if _debug:
                    ut.cprint('-------------', 'blue')
                if output_node.tablename != target_tablename:
                    # Get table configuration
                    config_ = depc._ensure_config(tablekey, config, _debug)

                    output_rowids = table.get_rowid(
                        _parent_rowids, config=config_, recompute=_recompute, **_kwargs
                    )
                    rowid_dict[output_node] = output_rowids
                    # table.get_model_inputs(table.get_model_uuid(output_rowids)[0])
                else:
                    # We are only computing up to the parents of the table here.
                    parent_rowids = _parent_rowids
                    break
            # rowids = rowid_dict[output_node]
            return parent_rowids

    def check_rowids(depc, tablename, input_tuple, config={}):
        """
        Returns a list of flags where True means the row has been computed and
        False means that it needs to be computed.
        """
        existing_rowids = depc.get_rowids(
            tablename, input_tuple, config=config, ensure=False
        )
        flags = ut.flag_not_None_items(existing_rowids)
        return flags

    def get_rowids(depc, tablename, input_tuple, **rowid_kw):
        """
        Used to get tablename rowids. Ensures rows exist unless ensure=False.
        rowids uniquely specify parent inputs and a configuration.

        CommandLine:
            python -m dtool.depcache_control get_rowids --show
            python -m dtool.depcache_control get_rowids:1

        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.dtool.depcache_control import *  # NOQA
            >>> from wbia.dtool.example_depcache2 import *  # NOQA
            >>> depc = testdata_depc3(True)
            >>> exec(ut.execstr_funckw(depc.get), globals())
            >>> kwargs = {}
            >>> root_rowids = [1, 2, 3]
            >>> root_rowids2 = [(4, 5, 6, 7)]
            >>> root_rowids3 = root_rowids2
            >>> _debug = True
            >>> tablename = 'smk_match'
            >>> input_tuple = (root_rowids, root_rowids2, root_rowids3)
            >>> target_table = depc[tablename]
            >>> inputs = target_table.rootmost_inputs.total_expand()
            >>> depc.get_rowids(tablename, input_tuple, _debug=_debug)
            >>> depc.print_all_tables()

        Example:
            >>> # ENABLE_DOCTEST
            >>> # Test external / ensure getters
            >>> from wbia.dtool.example_depcache import *  # NOQA
            >>> config = {}
            >>> depc = testdata_depc()
            >>> aids = [1,]
            >>> depc.delete_property('keypoint', aids, config=config)
            >>> chip_fpaths = depc.get('chip', aids, 'chip', config=config, read_extern=False)
            >>> ut.remove_file_list(chip_fpaths)
            >>> rowids = depc.get_rowids('keypoint', aids, ensure=True, config=config)
            >>> print('rowids = %r' % (rowids,))

        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.dtool.example_depcache import *  # NOQA
            >>> depc = testdata_depc()
            >>> depc.clear_all()
            >>> root_rowids = [1, 2]
            >>> config = {}
            >>> # Recompute the first few, make sure the rowids do not change
            >>> _ = depc.get_rowids('chip', root_rowids + [3], config=config)
            >>> assert _ == [1, 2, 3]
            >>> initial_rowids = depc.get_rowids('chip', root_rowids, config=config)
            >>> recomp_rowids = depc.get_rowids('chip', root_rowids, config=config, recompute=True)
            >>> assert recomp_rowids == initial_rowids, 'rowids should not change due to recompute'
        """
        target_tablename = tablename
        _debug = rowid_kw.get('_debug', False)
        _debug = depc._debug if _debug is None else _debug
        _kwargs = rowid_kw.copy()
        config = _kwargs.pop('config', {})
        _hack_rootmost = _kwargs.pop('_hack_rootmost', False)
        _recompute_all = _kwargs.pop('recompute_all', False)
        recompute = _kwargs.pop('recompute', _recompute_all)
        table = depc[target_tablename]

        parent_rowids = depc.get_parent_rowids(
            target_tablename,
            input_tuple,
            config=config,
            _hack_rootmost=_hack_rootmost,
            **_kwargs,
        )

        with ut.Indenter('[GetRowId-%s]' % (target_tablename,), enabled=_debug):
            config_ = depc._ensure_config(target_tablename, config, _debug)
            rowids = table.get_rowid(
                parent_rowids, config=config_, recompute=recompute, **_kwargs
            )
        return rowids

    @ut.accepts_scalar_input2(argx_list=[1])
    def get(
        depc,
        tablename,
        root_rowids,
        colnames=None,
        config=None,
        ensure=True,
        _debug=None,
        recompute=False,
        recompute_all=False,
        eager=True,
        nInput=None,
        read_extern=True,
        onthefly=False,
        num_retries=1,
        hack_paths=False,
    ):
        r"""
        Access dependant properties the primary objects using primary ids.

        Gets the data in `colnames` of `tablename` that correspond to
        `root_rowids` using `config`.  if colnames is None, all columns are
        returned.

        Args:
            tablename (str): table name containing desired property
            root_rowids (List[int]): ids of the root object
            colnames (None): desired property (default = None)
            config (None): (default = None)
            read_extern: if False then only returns extern URI
            hack_paths: if False then does not compute extern info just returns
                path that it will be located at

        Returns:
            list: prop_list

        CommandLine:
            python -m dtool.depcache_control --exec-get

        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.dtool.depcache_control import *  # NOQA
            >>> from wbia.dtool.example_depcache2 import *  # NOQA
            >>> from wbia.dtool.example_depcache import *  # NOQA
            >>> depc = testdata_depc3(True)
            >>> exec(ut.execstr_funckw(depc.get), globals())
            >>> aids = [1, 2, 3]
            >>> _debug = True
            >>> tablename = 'labeler'
            >>> root_rowids = aids
            >>> prop_list = depc.get(
            >>>     tablename, root_rowids, colnames)
            >>> result = ('prop_list = %s' % (ut.repr2(prop_list),))
            >>> print(result)
            prop_list = [('labeler([root(1)]:42)',), ('labeler([root(2)]:42)',), ('labeler([root(3)]:42)',)]

        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.dtool.depcache_control import *  # NOQA
            >>> from wbia.dtool.example_depcache2 import *  # NOQA
            >>> from wbia.dtool.example_depcache import *  # NOQA
            >>> depc = testdata_depc3(True)
            >>> exec(ut.execstr_funckw(depc.get), globals())
            >>> aids = [1, 2, 3]
            >>> _debug = True
            >>> tablename = 'smk_match'
            >>> tablename = 'vocab'
            >>> table = depc[tablename]
            >>> root_rowids = [aids]
            >>> prop_list = depc.get(
            >>>     tablename, root_rowids, colnames, config)
            >>> result = ('prop_list = %s' % (ut.repr2(prop_list),))
            >>> print(result)
            prop_list = [('vocab([root(1;2;3)]:42)',)]

        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.dtool.depcache_control import *  # NOQA
            >>> from wbia.dtool.example_depcache2 import *  # NOQA
            >>> from wbia.dtool.example_depcache import *  # NOQA
            >>> depc = testdata_depc3(True)
            >>> exec(ut.execstr_funckw(depc.get), globals())
            >>> aids = [1, 2, 3]
            >>> _debug = True
            >>> depc = testdata_depc()
            >>> tablename = 'chip'
            >>> table = depc[tablename]
            >>> root_rowids = aids
            >>> # Ensure chips are computed
            >>> prop_list1 = depc.get(tablename, root_rowids)
            >>> # Get file paths and delete them
            >>> prop_list2 = depc.get(tablename, root_rowids, read_extern=False)
            >>> n = ut.remove_file_list(ut.take_column(prop_list2, 1))
            >>> assert n == len(prop_list2), 'files were not computed'
            >>> prop_list3 = depc.get(tablename, root_rowids)
            >>> assert np.all(prop_list1[0][1] == prop_list3[0][1]), 'computed same info'
        """
        if tablename == depc.root_tablename:
            return depc.root_getters[colnames](root_rowids)
            # pass
        _debug = depc._debug if _debug is None else _debug
        with ut.Indenter('[GetProp-%s]' % (tablename,), enabled=_debug):
            if _debug:
                logger.info(' * tablename=%s' % (tablename))
                logger.info(' * root_rowids=%s' % (ut.trunc_repr(root_rowids)))
                logger.info(' * colnames = %r' % (colnames,))
                logger.info(' * config = %r' % (config,))

            if hack_paths and not ensure and not read_extern:
                # HACK: should be able to not compute rows to get certain properties
                from os.path import join

                # recompute_ = recompute or recompute_all
                parent_rowids = depc.get_parent_rowids(
                    tablename,
                    root_rowids,
                    config=config,
                    ensure=True,
                    _debug=None,
                    recompute_all=False,
                    eager=True,
                    nInput=None,
                )
                config_ = depc._ensure_config(tablename, config)
                if _debug:
                    logger.info(' * (ensured) config_ = %r' % (config_,))
                table = depc[tablename]
                extern_dpath = table.extern_dpath
                ut.ensuredir(extern_dpath, verbose=False or table.depc._debug)
                fname_list = table.get_extern_fnames(
                    parent_rowids, config=config_, extern_col_index=0
                )
                fpath_list = [join(extern_dpath, fname) for fname in fname_list]
                return fpath_list

            if nInput is None and ut.is_listlike(root_rowids):
                nInput = len(root_rowids)

            rowid_kw = dict(
                config=config,
                nInput=nInput,
                eager=eager,
                ensure=ensure,
                recompute=recompute,
                recompute_all=recompute_all,
                _debug=_debug,
            )

            rowdata_kw = dict(
                read_extern=read_extern,
                _debug=_debug,
                num_retries=num_retries,
                eager=eager,
                ensure=ensure,
                nInput=nInput,
            )

            input_tuple = root_rowids

            for trynum in range(num_retries + 1):
                try:
                    table = depc[tablename]
                    # Vectorized get of properties
                    tbl_rowids = depc.get_rowids(tablename, input_tuple, **rowid_kw)
                    if _debug:
                        logger.info(
                            '[depc.get] tbl_rowids = %s' % (ut.trunc_repr(tbl_rowids),)
                        )
                    prop_list = table.get_row_data(tbl_rowids, colnames, **rowdata_kw)
                except depcache_table.ExternalStorageException:
                    logger.info('!!* Hit ExternalStorageException')
                    if trynum == num_retries:
                        raise
                else:
                    break
            if _debug:
                logger.info('* return prop_list=%s' % (ut.trunc_repr(prop_list),))
        return prop_list

    def get_native(
        depc, tablename, tbl_rowids, colnames=None, _debug=None, read_extern=True
    ):
        """
        Gets data using internal ids, which is faster if you have them.

        CommandLine:
            python -m dtool.depcache_control get_native:0
            python -m dtool.depcache_control get_native:1

        Example:
            >>> # ENABLE_DOCTEST
            >>> # Simple test of get native
            >>> from wbia.dtool.example_depcache import *  # NOQA
            >>> config = {}
            >>> depc = testdata_depc()
            >>> tablename = 'keypoint'
            >>> aids = [1,]
            >>> tbl_rowids = depc.get_rowids(tablename, aids, config=config)
            >>> data = depc.get_native(tablename, tbl_rowids)

        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.dtool.example_depcache import *  # NOQA
            >>> depc = testdata_depc()
            >>> config = {}
            >>> tablename = 'chip'
            >>> colnames = extern_colname = 'chip'
            >>> aids = [1, 2]
            >>> depc.delete_property(tablename, aids, config=config)
            >>> # Ensure chip rowids exist then delete external data without
            >>> # notifying the depcache. This forces the depcache to recover
            >>> tbl_rowids = chip_rowids = depc.get_rowids(tablename, aids, config=config)
            >>> data_fpaths = depc.get(tablename, aids, extern_colname, config=config, read_extern=False)
            >>> ut.remove_file_list(data_fpaths)
            >>> chips = depc.get_native(tablename, tbl_rowids, extern_colname)
            >>> print('chips = %r' % (chips,))
        """
        tbl_rowids = list(tbl_rowids)
        _debug = depc._debug if _debug is None else _debug
        with ut.Indenter('[GetNative %s]' % (tablename,), enabled=_debug):
            if _debug:
                logger.info(' * tablename = %r' % (tablename,))
                logger.info(' * colnames = %r' % (colnames,))
                logger.info(' * tbl_rowids=%s' % (ut.trunc_repr(tbl_rowids)))
            table = depc[tablename]
            # import utool
            # with utool.embed_on_exception_context:
            # try:
            prop_list = table.get_row_data(
                tbl_rowids, colnames, _debug=_debug, read_extern=read_extern
            )
            # except depcache_table.ExternalStorageException:
            #    # This code is a bit rendant and would probably live better elsewhere
            #    # Also need to fix issues if more than one column specified
            #    extern_uris = table.get_row_data(
            #        tbl_rowids, colnames, _debug=_debug, read_extern=False,
            #        delete_on_fail=True, ensure=False)
            #    from os.path import exists
            #    error_flags = [exists(uri) for uri in extern_uris]
            #    redo_rowids = ut.compress(tbl_rowids, ut.not_list(error_flags))
            #    parent_rowids = table.get_parent_rowids(redo_rowids)
            #    # config_rowids = table.get_row_cfgid(redo_rowids)
            #    configs = table.get_row_configs(redo_rowids)
            #    assert ut.allsame(list(map(id, configs))), 'more than one config not yet supported'
            #    config = configs[0]
            #    table.get_rowid(parent_rowids, recompute=True, config=config)

            #    # TRY ONE MORE TIME
            #    prop_list = table.get_row_data(tbl_rowids, colnames, _debug=_debug,
            #                                   read_extern=read_extern,
            #                                   delete_on_fail=False)
        return prop_list

    def get_config_history(depc, tablename, root_rowids, config=None):
        # Vectorized get of properties
        tbl_rowids = depc.get_rowids(tablename, root_rowids, config=config)
        return depc[tablename].get_config_history(tbl_rowids)

    def get_root_rowids(depc, tablename, native_rowids):
        r"""
        Args:
            tablename (str):
            native_rowids (list):

        Returns:
            list:

        CommandLine:
            python -m dtool.depcache_control get_root_rowids --show

        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.dtool.example_depcache import *  # NOQA
            >>> depc = testdata_depc()
            >>> config1 = {'adapt_shape': False}
            >>> config2 = {'adapt_shape': True}
            >>> root_rowids = [2, 3, 5, 7]
            >>> native_rowids1 = depc.get_rowids('keypoint', root_rowids, config=config1)
            >>> native_rowids2 = depc.get_rowids('keypoint', root_rowids, config=config2)
            >>> ancestor_rowids1 = list(depc.get_root_rowids('keypoint', native_rowids1))
            >>> ancestor_rowids2 = list(depc.get_root_rowids('keypoint', native_rowids2))
            >>> assert native_rowids1 != native_rowids2, 'should have different native rowids'
            >>> assert ancestor_rowids1 == root_rowids, 'should have same root'
            >>> assert ancestor_rowids2 == root_rowids, 'should have same root'
        """
        return depc.get_ancestor_rowids(tablename, native_rowids, depc.root)

    def get_ancestor_rowids(depc, tablename, native_rowids, ancestor_tablename=None):
        """
        ancestor_tablename = depc.root; native_rowids = cid_list; tablename = const.CHIP_TABLE
        """
        if ancestor_tablename is None:
            ancestor_tablename = depc.root
        table = depc[tablename]
        ancestor_rowids = table.get_ancestor_rowids(native_rowids, ancestor_tablename)
        return ancestor_rowids

    def new_request(depc, tablename, qaids, daids, cfgdict=None):
        """ creates a request for data that can be executed later """
        logger.info('[depc] NEW %s request' % (tablename,))
        requestclass = depc.requestclass_dict[tablename]
        request = requestclass.new(depc, qaids, daids, cfgdict, tablename=tablename)
        return request

    # -----------------------------
    # STATE MODIFIERS

    def delete_property(depc, tablename, root_rowids, config=None, _debug=False):
        """
        Deletes the rowids of `tablename` that correspond to `root_rowids`
        using `config`.

        FIXME: make this work for all configs
        """
        rowid_list = depc.get_rowids(
            tablename, root_rowids, config=config, ensure=False, _debug=_debug
        )
        table = depc[tablename]
        num_deleted = table.delete_rows(rowid_list)
        return num_deleted

    def delete_property_all(depc, tablename, root_rowids, _debug=False):
        """
        Deletes the rowids of `tablename` that correspond to `root_rowids`
        using `config`.

        FIXME: make this work for all configs
        """
        table = depc[tablename]
        all_rowid_list = table._get_all_rowids()
        if len(all_rowid_list) == 0:
            return 0

        ancestor_rowid_list = depc.get_ancestor_rowids(tablename, all_rowid_list)

        rowid_list = []
        root_rowids_set = set(root_rowids)
        for rowid, ancestor_rowid in zip(all_rowid_list, ancestor_rowid_list):
            if ancestor_rowid in root_rowids_set:
                rowid_list.append(rowid)

        num_deleted = table.delete_rows(rowid_list)
        return num_deleted


@six.add_metaclass(ut.ReloadingMetaclass)
class DependencyCache(_CoreDependencyCache, ut.NiceRepr):
    """
    Currently, to use this class a user must:
        * on root modification, call depc.on_root_modified
        * use decorators to register relevant functions
    """

    def __init__(
        depc,
        root_tablename=None,
        cache_dpath='./DEPCACHE',
        controller=None,
        default_fname=None,
        # root_asobject=None,
        get_root_uuid=None,
        root_getters=None,
        use_globals=True,
    ):
        if default_fname is None:
            default_fname = root_tablename + '_primary_cache'
            # default_fname = ':memory:'
        depc.root_getters = root_getters
        # Root of all dependencies
        depc.root_tablename = root_tablename
        # Directory all cachefiles are stored in
        depc.cache_dpath = ut.truepath(cache_dpath)
        # Parent (ibs) controller
        depc.controller = controller
        # Internal dictionary of dependant tables
        depc.cachetable_dict = {}
        depc.configclass_dict = {}
        depc.requestclass_dict = {}
        depc.resultclass_dict = {}
        # Mapping of different files properties are stored in
        depc.fname_to_db = {}
        # Function to map a root rowid to an object
        # depc._root_asobject = root_asobject
        depc._use_globals = use_globals
        depc.default_fname = default_fname
        if get_root_uuid is None:
            logger.info('WARNING NEED UUID FUNCTION')
            # HACK
            get_root_uuid = ut.identity
        depc.get_root_uuid = get_root_uuid
        depc.delete_exclude_tables = {}
        depc._debug = ut.get_argflag(('--debug-depcache', '--debug-depc'))

    def get_tablenames(depc):
        return list(depc.cachetable_dict.keys())

    @property
    def tables(depc):
        return list(depc.cachetable_dict.values())

    @property
    def tablenames(depc):
        return depc.get_tablenames()

    def print_schemas(depc):
        for fname, db in depc.fname_to_db.items():
            logger.info('fname = %r' % (fname,))
            db.print_schema()

    # def print_table_csv(depc, tablename):
    #    depc[tablename]

    def print_table(depc, tablename):
        depc[tablename].print_table()

    def print_all_tables(depc):
        for tablename, table in depc.cachetable_dict.items():
            table.print_table()
            # db = table.db
            # db.print_table_csv(tablename)

    def print_config_tables(depc):
        for fname in depc.fname_to_db:
            logger.info('---')
            logger.info('db_fname = %r' % (fname,))
            depc.fname_to_db[fname].print_table_csv('config')

    def get_edges(depc, data=False):
        """
        edges for networkx structure
        """
        if data:

            def get_edgedata(tablekey, parentkey, parent_data):
                if parent_data['ismulti'] or parent_data['isnwise']:
                    edge_type_parts = []
                    # local_input_id = ''
                    if parent_data['ismulti']:
                        # TODO: give different ids to multi edges
                        # edge_type_parts.append('multi_%s_%s' % (parentkey, tablekey))
                        edge_type_parts.append('multi')
                        # local_input_id += '*'
                    if parent_data['isnwise']:
                        args = (parent_data['nwise_idx'],)
                        edge_type_parts.append('nwise_%s' % args)
                        # local_input_id += six.text_type(parent_data['nwise_idx'])
                        # edge_type_parts.append('nwise_%s_%s_%s' % (
                        #     parentkey, tablekey, parent_data['nwise_idx'],))
                    edge_type_id = '_'.join(edge_type_parts)
                else:
                    edge_type_id = 'normal'
                    # local_input_id = '1'
                edge_data = {
                    'ismulti': parent_data['ismulti'],
                    'isnwise': parent_data.get('isnwise'),
                    'nwise_idx': parent_data.get('nwise_idx'),
                    'parent_colx': parent_data.get('parent_colx'),
                    'edge_type': edge_type_id,
                    # 'local_input_id': local_input_id,
                    'local_input_id': parent_data.get('local_input_id'),
                    'taillabel': parent_data.get('local_input_id'),
                    # 'taillabel': local_input_id,  # proper graphviz attribute
                }
                return edge_data

            edges = [
                (parentkey, tablekey, get_edgedata(tablekey, parentkey, parent_data))
                for tablekey, table in depc.cachetable_dict.items()
                for parentkey, parent_data in table.parents(data=True)
            ]
        else:
            edges = [
                (parentkey, tablekey)
                for tablekey, table in depc.cachetable_dict.items()
                for parentkey in table.parents(data=False)
            ]
        return edges

    def get_implicit_edges(depc, data=False):
        """
        Edges defined by subconfigurations
        """
        # add implicit edges
        implicit_edges = []
        # Map config classes to tablenames
        _inverted_ccdict = ut.invert_dict(depc.configclass_dict)
        for tablename2, configclass in depc.configclass_dict.items():
            cfg = configclass()
            subconfigs = cfg.get_sub_config_list()
            if subconfigs is not None and len(subconfigs) > 0:
                tablename1_list = ut.dict_take(_inverted_ccdict, subconfigs, None)
                for tablename1 in ut.filter_Nones(tablename1_list):
                    implicit_edges.append((tablename1, tablename2))
        if data:
            implicit_edges = [(e1, e2, {'implicit': True}) for e1, e2 in implicit_edges]
        return implicit_edges

    @ut.memoize
    def make_graph(depc, **kwargs):
        """
        Constructs a networkx representation of the dependency graph

        CommandLine:
            python -m dtool --tf DependencyCache.make_graph --show --reduced

            python -m wbia.control.IBEISControl show_depc_annot_graph --show --reduced

            python -m wbia.control.IBEISControl show_depc_annot_graph --show --reduced --testmode
            python -m wbia.control.IBEISControl show_depc_annot_graph --show --testmode

            python -m wbia.control.IBEISControl --test-show_depc_image_graph --show --reduced
            python -m wbia.control.IBEISControl --test-show_depc_image_graph --show

            python -m wbia.scripts.specialdraw double_depcache_graph --show --testmode

        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.dtool.depcache_control import *  # NOQA
            >>> from wbia.dtool.example_depcache import testdata_depc
            >>> import utool as ut
            >>> depc = testdata_depc()
            >>> graph = depc.make_graph(reduced=ut.get_argflag('--reduced'))
            >>> ut.quit_if_noshow()
            >>> import wbia.plottool as pt
            >>> pt.ensureqt()
            >>> import networkx as nx
            >>> #pt.show_nx(nx.dag.transitive_closure(graph))
            >>> #pt.show_nx(ut.nx_transitive_reduction(graph))
            >>> pt.show_nx(graph)
            >>> pt.show_nx(graph, layout='agraph')
            >>> ut.show_if_requested()

        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.dtool.depcache_control import *  # NOQA
            >>> from wbia.dtool.example_depcache import testdata_depc
            >>> import utool as ut
            >>> depc = testdata_depc()
            >>> graph = depc.make_graph(reduced=True)
            >>> # xdoctest: +REQUIRES(--show)
            >>> ut.quit_if_noshow()
            >>> import wbia.plottool as pt
            >>> pt.ensureqt()
            >>> import networkx as nx
            >>> #pt.show_nx(nx.dag.transitive_closure(graph))
            >>> #pt.show_nx(ut.nx_transitive_reduction(graph))
            >>> pt.show_nx(graph)
            >>> pt.show_nx(graph, layout='agraph')
            >>> ut.show_if_requested()
        """
        import networkx as nx

        # graph = nx.DiGraph()
        graph = nx.MultiDiGraph()
        nodes = list(depc.cachetable_dict.keys())
        edges = depc.get_edges(data=True)
        graph.add_nodes_from(nodes)
        graph.add_edges_from(edges)

        if kwargs.get('implicit', True):
            implicit_edges = depc.get_implicit_edges(data=True)
            graph.add_edges_from(implicit_edges)

        shape_dict = {
            # 'node': 'circle',
            # 'node': 'rect',
            'node': 'ellipse',
            # 'root': 'rhombus',
            # 'root': 'circle',
            # 'root': 'circle',
            'root': 'ellipse',
            # 'root': 'rect',
        }
        # import wbia.plottool as pt
        NEUTRAL_BLUE = np.array((159, 159, 241, 255)) / 255.0
        RED = np.array((255, 0, 0, 255)) / 255.0
        color_dict = {
            # 'algo': pt.DARK_GREEN,  # 'g',
            'node': NEUTRAL_BLUE,
            'root': RED,  # 'r',
        }

        def _node_attrs(dict_):
            props = {k: dict_['node'] for k, v in depc.cachetable_dict.items()}
            props[depc.root] = dict_['root']
            return props

        nx.set_node_attributes(graph, name='color', values=_node_attrs(color_dict))
        nx.set_node_attributes(graph, name='shape', values=_node_attrs(shape_dict))
        if kwargs.get('reduced', False):
            # FIXME; There is a bug in the reduction of the image depc graph
            # Reduce only the non-multi part of the graph
            nonmulti_graph = graph.copy()
            multi_data_edges = [
                (u, v, d) for u, v, d in graph.edges(data=True) if d.get('ismulti')
            ]
            multi_edges = [(u, v) for u, v, d in multi_data_edges]
            nonmulti_graph.remove_edges_from(multi_edges)

            # Hack to recognize that implicit edges on multi-input edges
            # PROBABLY means that the implicit edge is also a multi edge This
            # needs to be fixed by indicating which parent edge the implicit
            # edge actually corresponds to.
            multi_data_edges_ = []
            for edge in multi_data_edges:
                node = edge[1]
                in_edges = list(nonmulti_graph.in_edges(node, data=True))
                if len(in_edges) == 1:
                    u, v, d = in_edges[0]
                    # If there is only one implicit edge on a multi-edge, it very
                    # likely means that the implicit edge should have the same
                    # input-id as the multi-edge
                    if d.get('implicit'):
                        nonmulti_graph.remove_edge(u, v)
                    # hack the multi-edge to reduce to the implicit version
                    new_edge = (u, edge[1], edge[2])
                    multi_data_edges_.append(new_edge)
                else:
                    multi_data_edges_.append(edge)
            multi_data_edges = multi_data_edges_

            # <ATTEMPT>
            # The transitive reduction should respect impolicit edges,
            # but the end result should not contain any implicit edges

            # Transitive reduction wrt implicit edges
            # For every node...
            implicit_aware = 1

            if implicit_aware:
                removed_in_edges = {}
                for node in graph.nodes():
                    in_edges = list(graph.in_edges(node, data=True))
                    # if there is an implicit incoming edge
                    implicit_flags = [edge[2].get('implicit') for edge in in_edges]
                    explicit_flags = ut.not_list(implicit_flags)
                    implicit_edges = ut.compress(in_edges, implicit_flags)
                    flag = True
                    for edge in implicit_edges:
                        # Ignore this edge if there is a common descendant
                        if ut.nx_common_descendants(graph, node, edge[0]):
                            pass
                            # removed_in_edges[node] = implicit_edges
                            # flag = False

                    if flag and any(implicit_edges):
                        # then remove all non-implicit incoming edges
                        remove_non_implicit = ut.compress(in_edges, explicit_flags)
                        # remember this nodes removed in edges
                        removed_in_edges[node] = remove_non_implicit
                to_remove2 = ut.take_column(ut.flatten(removed_in_edges.values()), [0, 1])
                nonmulti_graph.remove_edges_from(to_remove2)

            graph_tr = ut.nx_transitive_reduction(nonmulti_graph)

            if implicit_aware:
                # Place old non-implicit structure on top of reduced implicit edges
                for node in removed_in_edges.keys():
                    old_edges = removed_in_edges[node]
                    for edge in graph_tr.in_edges(node, data=True):
                        data = edge[2]
                        for old_edge in old_edges:
                            # TODO: handle multiple old edges
                            data.update(old_edge[2])
                            data['implicit'] = False

            # HACK IN STRUCTURE
            # Multi Edges
            # (doesn't quite work)
            if False:
                for u, v, data in graph.edges(data=True):
                    if data.get('ismulti'):
                        new_parent = nx.shortest_path(graph_tr, u, v)[-2]
                        # graph_tr[new_parent][v][0]['is_multi'] = True
                        logger.info('NEW MULTI')
                        logger.info((new_parent, v))
                        nx.set_edge_attributes(
                            graph_tr, name='ismulti', values={(new_parent, v, 0): True}
                        )
                        # logger.info(v)
            else:
                pass
                graph_tr.add_edges_from(multi_data_edges)

            parents = ut.ddict(list)
            for u, v, data in graph.edges(data=True):
                parents[v].append(u)

            for node, ps in parents.items():
                num_connect = ut.dict_hist(ps)
                nwise_parents = [(k, v) for k, v in num_connect.items() if v > 1]

                for p, n in nwise_parents:
                    new_parent = nx.shortest_path(graph_tr, p, node)[-2]
                    for x in range(n - 1):
                        # import utool
                        # utool.embed()
                        graph_tr.add_edge(new_parent, node)
                    # G_tr[new_parent][v][0]['is_multi'] = True
            nx.set_node_attributes(graph_tr, name='color', values=_node_attrs(color_dict))
            nx.set_node_attributes(graph_tr, name='shape', values=_node_attrs(shape_dict))
            graph = graph_tr

        if kwargs.get('remove_local_input_id', False):
            ut.nx_delete_edge_attr(graph, 'local_input_id')

        return graph

    @property
    def graph(depc):
        return depc.make_graph()

    @property
    def explicit_graph(depc):
        return depc.make_graph(implicit=False)

    @property
    def reduced_graph(depc):
        return depc.make_graph(reduced=True)

    def show_graph(depc, reduced=False, **kwargs):
        """ Helper "fluff" function """
        import wbia.plottool as pt

        graph = depc.make_graph(reduced=reduced)
        if ut.is_developer():
            ut.ensureqt()
        kwargs['layout'] = 'agraph'
        pt.show_nx(graph, **kwargs)

    def __nice__(depc):
        infostr_ = 'nTables=%d' % len(depc.cachetable_dict)
        return '(%s) %s' % (depc.root_tablename, infostr_)

    def __getitem__(depc, tablekey):
        return depc.cachetable_dict[tablekey]

    @property
    def root(depc):
        return depc.root_tablename

    def delete_root(
        depc,
        root_rowids,
        delete_extern=None,
        _debug=False,
        table_config_filter=None,
        prop=None,
    ):
        r"""
        Deletes all properties of a root object regardless of config

        Args:
            root_rowids (list):

        CommandLine:
            python -m dtool.depcache_control delete_root --show

        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.dtool.depcache_control import *  # NOQA
            >>> from wbia.dtool.example_depcache import testdata_depc
            >>> depc = testdata_depc()
            >>> exec(ut.execstr_funckw(depc.delete_root), globals())
            >>> root_rowids = [1]
            >>> depc.delete_root(root_rowids, _debug=0)
            >>> depc.get('fgweight', [1])
            >>> depc.delete_root(root_rowids, _debug=0)
        """
        # graph = depc.make_graph(implicit=False)
        # hack
        # check to make sure child does not have another parent
        rowid_dict = depc.get_allconfig_descendant_rowids(
            root_rowids, table_config_filter
        )
        # children = [child for child in graph.succ[depc.root_tablename]
        #            if sum([len(e) for e in graph.pred[child].values()]) == 1]
        # depc.delete_property(tablename, root_rowids, _debug=_debug)
        num_deleted = 0
        for tablename, table_rowids in rowid_dict.items():
            if tablename == depc.root:
                continue
            # Specific prop exclusion
            delete_exclude_table_set_prop = depc.delete_exclude_tables.get(prop, [])
            delete_exclude_table_set_all = depc.delete_exclude_tables.get(None, [])
            if (
                tablename in delete_exclude_table_set_prop
                or tablename in delete_exclude_table_set_all
            ):
                continue
            table = depc[tablename]
            num_deleted += table.delete_rows(table_rowids, delete_extern=delete_extern)
        return num_deleted

    def register_delete_table_exclusion(depc, tablename, prop):
        if prop not in depc.delete_exclude_tables:
            depc.delete_exclude_tables[prop] = set([])
        depc.delete_exclude_tables[prop].add(tablename)
        args = (ut.repr3(depc.delete_exclude_tables),)
        logger.info('[depc] Updated delete tables: %s' % args)

    def get_allconfig_descendant_rowids(depc, root_rowids, table_config_filter=None):
        import networkx as nx

        # list(nx.topological_sort(nx.bfs_tree(graph, depc.root)))
        # decendants = nx.descendants(graph, depc.root)
        # raise NotImplementedError()

        graph = depc.make_graph(implicit=True)
        root = depc.root
        rowid_dict = {}
        rowid_dict[root] = root_rowids

        # Find all rowids that inherit from the specific root rowids
        sinks = list(ut.nx_sink_nodes(nx.bfs_tree(graph, depc.root)))
        for target_tablename in sinks:
            path = nx.shortest_path(graph, root, target_tablename)
            for parent, child in ut.itertwo(path):
                child_table = depc[child]
                relevant_col_attrs = [
                    attrs
                    for attrs in child_table.parent_col_attrs
                    if attrs['parent_table'] == parent
                ]
                parent_colnames = [
                    attrs['intern_colname'] for attrs in relevant_col_attrs
                ]

                params_iter = list(zip(rowid_dict[parent]))

                for parent_colname in parent_colnames:
                    child_rowids = child_table.db.get_where_eq_set(
                        child_table.tablename,
                        (child_table.rowid_colname,),
                        params_iter,
                        unpack_scalars=False,
                        where_colnames=[parent_colname],
                        op='AND',
                    )
                    # child_rowids = ut.flatten(child_rowids)
                    if table_config_filter is not None:
                        config_filter = table_config_filter.get(
                            child_table.tablename, None
                        )
                        if config_filter is not None:
                            # If a config filter is specified only grab rows that
                            # meed the filter
                            tbl_cfgids = child_table.get_row_cfgid(child_rowids)
                            cfgid2_rowids = ut.group_items(child_rowids, tbl_cfgids)
                            unique_cfgids = cfgid2_rowids.keys()
                            unique_cfgids = ut.filter_Nones(unique_cfgids)
                            unique_configs = child_table.get_config_from_rowid(
                                unique_cfgids
                            )
                            passed_rowids = []
                            for config, cfgid in zip(unique_configs, tbl_cfgids):
                                if all(
                                    [
                                        config[key] == val
                                        for key, val in config_filter.items()
                                    ]
                                ):
                                    passed_rowids.extend(cfgid2_rowids[cfgid])
                            child_rowids = passed_rowids
                    rowid_dict[child] = ut.unique(
                        child_rowids + rowid_dict.get(child, [])
                    )
        return rowid_dict

    def notify_root_changed(depc, root_rowids, prop, force_delete=False):
        """
        this is where we are notified that a "registered" root property has
        changed.
        """
        logger.info(
            '[depc] notified that columns (%s) for (%d) row(s) were modified'
            % (prop, len(root_rowids))
        )
        # for key in tables_depending_on(prop)
        # depc.delete_property(key, root_rowids)
        # TODO: check which properties were invalidated by this prop
        # TODO; remove invalidated properties
        if force_delete:
            depc.delete_root(root_rowids, prop=prop)

    def clear_all(depc):
        logger.info('Clearning all cached data in %r' % (depc,))
        for table in depc.cachetable_dict.values():
            table.clear_table()

    def make_root_info_uuid(depc, root_rowids, info_props):
        """
        Creates a uuid that depends on certain properties of the root object.
        This is used for implicit cache invalidation because, if those
        properties change then this uuid also changes.

        The depcache needs to know about stateful properties of dynamic root
        objects in order to correctly compute their hashes.

        >>> #ibs = wbia.opendb(defaultdb='testdb1')
        >>> root_rowids = ibs._get_all_aids()
        >>> depc = ibs.depc_annot
        >>> info_props = ['image_uuid', 'verts', 'theta']
        >>> info_props = ['image_uuid', 'verts', 'theta', 'name', 'species', 'yaw']
        """
        getters = ut.dict_take(depc.root_getters, info_props)
        infotup_list = zip(*[getter(root_rowids) for getter in getters])
        info_uuid_list = [ut.augment_uuid(*tup) for tup in infotup_list]
        return info_uuid_list

    def get_uuids(depc, tablename, root_rowids, config=None):
        """
        # TODO: Make uuids for dependant object based on root uuid and path of
        # construction.
        """
        if tablename == depc.root:
            uuid_list = depc.get_root_uuid(root_rowids)
        return uuid_list

    get_native_property = _CoreDependencyCache.get_native
    get_property = _CoreDependencyCache.get

    def stacked_config(depc, source, dest, config):
        r"""
        CommandLine:
            python -m dtool.depcache_control stacked_config --show

        Example:
            >>> # ENABLE_DOCTEST
            >>> from wbia.dtool.depcache_control import *  # NOQA
            >>> from wbia.dtool.example_depcache import testdata_depc
            >>> depc = testdata_depc()
            >>> source = depc.root
            >>> dest = 'fgweight'
            >>> config = {}
            >>> stacked_config = depc.stacked_config(source, dest, config)
            >>> cfgstr = stacked_config.get_cfgstr()
            >>> result = ('cfgstr = %s' % (ut.repr2(cfgstr),))
            >>> print(result)
        """
        if config is None:
            config = {}
        if source is None:
            source = depc.root
        graph = depc.make_graph(implicit=True)
        requires_tables = ut.setdiff(
            ut.nx_all_nodes_between(graph, source, dest), [source]
        )
        # requires_tables = ut.setdiff(ut.nx_all_nodes_between(depc.graph, 'annotations', 'featweight'), ['annotations'])
        requires_tables = ut.nx_topsort_nodes(depc.graph, requires_tables)
        requires_configs = [
            depc.configclass_dict[tblname](**config) for tblname in requires_tables
        ]
        # cfgstr_list = [cfg.get_cfgstr() for cfg in requires_configs]
        stacked_config = base.StackedConfig(requires_configs)
        return stacked_config
        # cfgstr = stacked_config.get_cfgstr()
        # cfgstr = '_'.join(cfgstr_list)
        # return cfgstr



#! /usr/bin/env python
# -*- coding: utf-8 -*-


import os


def _kwargs(kwargs, key, value):
    if key not in kwargs.keys():
        kwargs[key] = value


def _parseFilename(path):
    name, ext = os.path.splitext(path)
    ext = ext[1:].lower()
    return name, ext


class Directory(object):
    def __init__(self, directory_path, **kwargs):
        _kwargs(kwargs, 'include_file_extensions', None)
        _kwargs(kwargs, 'include_hidden', False)
        _kwargs(kwargs, 'exclude_file_extensions', [])
        _kwargs(kwargs, 'recursive', False)
        _kwargs(kwargs, 'absolute', True)
        _kwargs(kwargs, 'image', None)
        _kwargs(kwargs, 'images', None)

        if kwargs['include_file_extensions'] == 'images' or True in [
            kwargs['image'],
            kwargs['images'],
        ]:
            kwargs['include_file_extensions'] = ['jpg', 'jpeg', 'png', 'tiff']

        if kwargs['absolute']:
            directory_path = os.path.abspath(os.path.expanduser(directory_path))

        if not os.path.exists(directory_path):
            raise Exception(
                'DIRECTORY_EXISTENCE',
                'The directory path you specified ['
                + directory_path
                + '] does not exist.',
            )

        self.directory_path = directory_path
        self.absolute_directory_path = os.path.abspath(directory_path)

        self.recursive = self._fix_recursive(kwargs['recursive'])
        kwargs['recursive'] = self.recursive

        self.file_list = []
        self.directory_list = []
        for line in os.listdir(self.absolute_directory_path):
            line_path = os.path.join(self.absolute_directory_path, line)
            filename, line_extension = _parseFilename(line_path)

            if (
                os.path.isfile(line_path)
                and (kwargs['include_hidden'] or line[0] != '.')
                and (
                    kwargs['include_file_extensions'] is None
                    or line_extension in kwargs['include_file_extensions']
                )
                and (line_extension not in kwargs['exclude_file_extensions'])
            ):
                self.file_list.append(line_path)
            elif (
                os.path.isdir(line_path)
                and kwargs['recursive'] >= 0
                and (kwargs['include_hidden'] or line[0] != '.')
            ):
                kwargs_ = kwargs.copy()
                kwargs_['recursive'] -= 1
                self.directory_list.append(Directory(line_path, **kwargs_))

    def __str__(self):
        return '<Directory Object | %s | %d files | %d directories>' % (
            self.absolute_directory_path,
            len(self.file_list),
            len(self.directory_list),
        )

    def __repr__(self):
        return '<Directory Object | ../%s>' % (self.base())

    def __iter__(self):
        for filename in self.files():
            yield filename

    def __lt__(direct1, direct2):
        if direct1.absolute_directory_path < direct2.absolute_directory_path:
            return -1
        if direct1.absolute_directory_path > direct2.absolute_directory_path:
            return 1
        return 0

    def _fix_recursive(self, recursive):
        if isinstance(recursive, bool):
            recursive = 10 ** 9 if recursive else -1
        else:
            recursive = int(recursive)

        assert isinstance(recursive, int)
        return recursive

    def base(self):
        return os.path.basename(self.absolute_directory_path)

    def files(self, **kwargs):
        _kwargs(kwargs, 'recursive', self.recursive)
        _kwargs(kwargs, 'absolute', False)

        kwargs['recursive'] = self._fix_recursive(kwargs['recursive'])

        directory_files = []
        if kwargs['recursive'] >= 0:
            for directory in self.directory_list:
                directory_files += directory.files(**kwargs)

        file_list = self.file_list
        if kwargs['absolute']:
            file_list = map(os.path.basename, file_list)
        return sorted(file_list + directory_files)

    def directories(self, **kwargs):
        _kwargs(kwargs, 'recursive', self.recursive)

        kwargs['recursive'] = self._fix_recursive(kwargs['recursive'])

        directory_dirs = []
        if kwargs['recursive'] >= 0:
            for directory in self.directory_list:
                directory_dirs += directory.directories(**kwargs)

        return sorted(self.directory_list + directory_dirs)

    def num_files(self, **kwargs):
        _kwargs(kwargs, 'recursive', self.recursive)

        kwargs['recursive'] = self._fix_recursive(kwargs['recursive'])

        directories_num_files = 0

        if kwargs['recursive'] >= 0:
            for directory in self.directory_list:
                directories_num_files += directory.num_files(**kwargs)

        return len(self.file_list) + directories_num_files

    def num_directories(self, **kwargs):
        _kwargs(kwargs, 'recursive', self.recursive)

        kwargs['recursive'] = self._fix_recursive(kwargs['recursive'])

        directories_num_directories = 0

        if kwargs['recursive'] >= 0:
            for directory in self.directory_list:
                directories_num_directories += directory.num_directories(**kwargs)

        return len(self.directory_list) + directories_num_directories


if __name__ == '__main__':
    # Test directory does not exist
    try:
        direct = Directory('test/')
    except Exception as e:
        assert e[0] == 'DIRECTORY_EXISTENCE'

    # Test recursive
    direct = Directory('tests/directory', recursive=True)
    assert direct.num_files(recursive=False) == 10
    assert direct.num_files() == 15
    assert direct.num_directories(recursive=False) == 2
    assert direct.num_directories() == 5

    # Test not recursive
    direct = Directory('tests/directory')
    assert direct.num_files(recursive=False) == 10
    assert direct.num_files() == 10
    assert direct.num_directories(recursive=False) == 0
    assert direct.num_directories() == 0

    # Test include
    direct = Directory('tests/directory', include_file_extensions=['txt'], recursive=True)
    assert direct.num_files(recursive=False) == 3
    assert direct.num_files() == 6

    # Test exclude
    direct = Directory(
        'tests/directory', exclude_file_extensions=['ignore'], recursive=True
    )
    assert direct.num_files(recursive=False) == 9
    assert direct.num_files(recursive=True) == 13



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
mkinit ~/code/wbia/wbia



DEV SCRIPT

TODO: DEPRICATE

This is a hacky script meant to be run mostly automatically with the option of
interactions.

dev.py is supposed to be a developer non-gui interface into the IBEIS software.
dev.py runs experiments and serves as a scratchpad for new code and quick scripts

TODO:

    Test to find typical "good" descriptor scores.  Find nearest neighbors and
    noramlizers for each feature in a query image.  Based on ground truth and
    spatial verification mark feature matches as true or false.  Visualize the
    feature scores of good matches vs bad matches. Lowe shows the pdf of
    correct matches and the PDF for incorrect matches. We should also show the
    same thing.

Done:
    Cache nearest neighbors so different parameters later in the pipeline dont
    take freaking forever.

CommandLine:
    python dev.py --wshow -t query --db PZ_MTEST --qaid 110 --cfg score_method:nsum prescore_method:nsum
    python dev.py --wshow -t query --db PZ_MTEST --qaid 110
    python dev.py --wshow -t query --db PZ_MTEST --qaid 110 --cfg fg_on=True
    python dev.py --wshow -t query --db PZ_MTEST --qaid 110 --cfg
"""
# TODO: ADD COPYRIGHT TAG
import logging
import sys
import numpy as np
from wbia._devscript import devcmd, DEVCMD_FUNCTIONS, DEVPRECMD_FUNCTIONS
import utool as ut
from utool.util_six import get_funcname
import utool

# from wbia.algo.hots import smk
import wbia.plottool as pt
import wbia

# import multiprocessing
# if __name__ == '__main__':
#     multiprocessing.freeze_support()
#     wbia._preload()
# utool.util_importer.dynamic_import(__name__, ('_devcmds_wbia', None),
#                                   developing=True)
from wbia._devscript import devcmd, devprecmd  # NOQA
from os.path import split, join, expanduser  # NOQA
from wbia.plottool import draw_func2 as df2  # NOQA
from wbia import sysres  # NOQA
from wbia.other import ibsfuncs  # NOQA
from wbia.dbio import ingest_hsdb  # NOQA
from wbia._devcmds_wbia import (  # NOQA
    GZ_VIEWPOINT_EXPORT_PAIRS,
    MOTHERS_VIEWPOINT_EXPORT_PAIRS,
    change_names,
    convert_hsdbs,
    delete_all_chips,
    delete_all_feats,
    delete_cache,
    ensure_mtest,
    ensure_nauts,
    ensure_wilddogs,
    export,
    list_dbs,
    list_unconverted_hsdbs,
    openworkdirs_test,
    query_aids,
    show_aids,
    sver_aids,
    vdd,
)

# IBEIS
from wbia.init import main_helpers  # NOQA
from wbia.other import dbinfo  # NOQA
from wbia.expt import experiment_configs  # NOQA
from wbia.expt import harness  # NOQA
from wbia.expt.experiment_drawing import draw_annot_scoresep
from wbia import params  # NOQA

print, rrr, profile = utool.inject2(__name__)
logger = logging.getLogger('wbia')


logger.setLevel(logging.INFO)


# ------------------
# DEV DEVELOPMENT
# ------------------
# This is where you write all of the functions that will become pristine
# and then go in _devcmds_wbia.py


"""
./dev.py -e print_results --db PZ_Master1 -a varysize_pzm:dper_name=[1,2],dsize=1500 -t candidacy_k:K=1 --intersect_hack
./dev.py -e draw_rank_cmc -t baseline -a baseline --show --db PZ_Master1
./dev.py -e get_dbinfo --db PZ_Master1 --aid_list=baseline
./dev.py -e get_dbinfo --db PZ_MTEST
./dev.py -e get_dbinfo --db PZ_Master1 --aid_list=baseline --hackshow-unixtime --show
./dev.py -e get_dbinfo --db PZ_Master1 --hackshow-unixtime --show
"""
# Quick interface into specific registered doctests
REGISTERED_DOCTEST_EXPERIMENTS = [
    (
        'wbia.expt.experiment_drawing',
        'draw_case_timedeltas',
        ['timedelta_hist', 'timedelta_pie'],
    ),
    ('wbia.expt.experiment_drawing', 'draw_match_cases', ['draw_cases', 'cases']),
    ('wbia.expt.experiment_drawing', 'draw_casetag_hist', ['taghist']),
    ('wbia.expt.old_storage', 'draw_results'),
    ('wbia.expt.experiment_drawing', 'draw_rank_cmc', ['rank_cmc']),
    ('wbia.other.dbinfo', 'get_dbinfo'),
    ('wbia.other.dbinfo', 'latex_dbstats'),
    ('wbia.other.dbinfo', 'show_image_time_distributions', ['db_time_hist']),
    ('wbia.expt.experiment_drawing', 'draw_rank_surface', ['rank_surface']),
    ('wbia.expt.experiment_helpers', 'get_annotcfg_list', ['print_acfg']),
    ('wbia.expt.experiment_printres', 'print_results', ['printres', 'print']),
    ('wbia.expt.experiment_printres', 'print_latexsum', ['latexsum']),
    ('wbia.dbio.export_subset', 'export_annots'),
    (
        'wbia.expt.experiment_drawing',
        'draw_annot_scoresep',
        ['scores', 'scores_good', 'scores_all'],
    ),
]


def _exec_doctest_func(modname, funcname):
    module = ut.import_modname(modname)
    func = module.__dict__[funcname]
    testsrc = ut.get_doctest_examples(func)[0][0]
    exec(testsrc, globals(), locals())


def _register_doctest_precmds():
    from functools import partial

    for tup in REGISTERED_DOCTEST_EXPERIMENTS:
        modname, funcname = tup[:2]
        aliases = tup[2] if len(tup) == 3 else []
        aliases += [funcname]
        _doctest_func = partial(_exec_doctest_func, modname, funcname)
        devprecmd(*aliases)(_doctest_func)


_register_doctest_precmds()


@devcmd('tune', 'autotune')
def tune_flann(ibs, qaid_list, daid_list=None):
    r"""
    CommandLine:
        python dev.py -t tune --db PZ_MTEST
        python dev.py -t tune --db GZ_ALL
        python dev.py -t tune --db GIR_Tanya
        python dev.py -t tune --db PZ_Master0
    """
    all_aids = ibs.get_valid_aids()
    vecs = np.vstack(ibs.get_annot_vecs(all_aids))
    print(
        'Tunning flann for species={species}:'.format(
            species=ibs.get_database_species(all_aids)
        )
    )
    import vtool as vt

    tuned_params = vt.tune_flann(
        vecs,
        target_precision=0.98,
        build_weight=0.05,
        memory_weight=0.00,
        sample_fraction=0.1,
    )
    tuned_params

    # tuned_params2 = vt.tune_flann(vecs,
    #                              target_precision=.90,
    #                              build_weight=0.001,
    #                              memory_weight=0.00,
    #                              sample_fraction=0.5)
    # tuned_params2


@devcmd('incremental', 'inc')
def incremental_test(ibs, qaid_list, daid_list=None):
    """
    Adds / queries new images one at a time to a clean test database.
    Tests the complete system.

    Args:
        ibs       (list) : IBEISController object
        qaid_list (list) : list of annotation-ids to query

    CommandLine:
        python dev.py -t inc --db PZ_MTEST --qaid 1:30:3 --cmd

        python dev.py --db PZ_MTEST --allgt --cmd

        python dev.py --db PZ_MTEST --allgt -t inc

        python dev.py -t inc --db PZ_MTEST --qaid 1:30:3 --cmd

        python dev.py -t inc --db GZ_ALL --ninit 100 --noqcache

        python dev.py -t inc --db PZ_MTEST --noqcache --interactive-after 40
        python dev.py -t inc --db PZ_Master0 --noqcache --interactive-after 10000 --ninit 400

    Example:
        >>> # DISABLE_DOCTEST
        >>> import wbia
        >>> ibs = wbia.opendb('PZ_MTEST')
        >>> qaid_list = ibs.get_valid_aids()
        >>> daid_list = None
    """
    from wbia.algo.hots import automated_matcher

    ibs1 = ibs
    num_initial = ut.get_argval('--ninit', type_=int, default=0)
    return automated_matcher.incremental_test(ibs1, num_initial)


@devcmd('inspect')
def inspect_matches(ibs, qaid_list, daid_list):
    print('<inspect_matches>')
    from wbia.gui import inspect_gui

    return inspect_gui.test_review_widget(ibs, qaid_list, daid_list)


def get_ibslist(ibs):
    print('[dev] get_ibslist')
    ibs_GV = ibs
    ibs_RI = ibs.clone_handle(nogravity_hack=True)
    ibs_RIW = ibs.clone_handle(nogravity_hack=True, gravity_weighting=True)
    ibs_list = [ibs_GV, ibs_RI, ibs_RIW]
    return ibs_list


@devcmd('gv_scores')
def compgrav_draw_score_sep(ibs, qaid_list, daid_list):
    print('[dev] compgrav_draw_score_sep')
    ibs_list = get_ibslist(ibs)
    for ibs_ in ibs_list:
        draw_annot_scoresep(ibs_, qaid_list)


# --------------------
# RUN DEV EXPERIMENTS
# --------------------


# def run_registered_precmd(precmd_name):
#    # Very hacky way to run just a single registered precmd
#    for (func_aliases, func) in DEVPRECMD_FUNCTIONS:
#        for aliases in func_aliases:
#            ret = precmd_name in input_precmd_list
#            if ret:
#                func()


def run_devprecmds():
    """
    Looks for pre-tests specified with the -t flag and runs them
    """
    # input_precmd_list = params.args.tests[:]
    input_precmd_list = ut.get_argval('-e', type_=list, default=[])
    valid_precmd_list = []

    def intest(*args, **kwargs):
        for precmd_name in args:
            valid_precmd_list.append(precmd_name)
            ret = precmd_name in input_precmd_list
            ret2 = precmd_name in params.unknown  # Let unparsed args count towards tests
            if ret or ret2:
                if ret:
                    input_precmd_list.remove(precmd_name)
                else:
                    ret = ret2
                print('+===================')
                print('| running precmd = %s' % (args,))
                return ret
        return False

    # Implicit (decorated) test functions
    for (func_aliases, func) in DEVPRECMD_FUNCTIONS:
        if intest(*func_aliases):
            # with utool.Indenter('[dev.' + get_funcname(func) + ']'):
            func()
            print('Exiting after first precommand')
            sys.exit(1)
    if len(input_precmd_list) > 0:
        raise AssertionError('Unhandled tests: ' + repr(input_precmd_list))


# @utool.indent_func('[dev]')
def run_devcmds(ibs, qaid_list, daid_list, acfg=None):
    """
    This function runs tests passed in with the -t flag
    """
    params.parse_args()
    print('\n')
    # print('[dev] run_devcmds')
    print('==========================')
    print('[DEV] RUN EXPERIMENTS %s' % ibs.get_dbname())
    print('==========================')
    input_test_list = params.args.tests[:]
    print('input_test_list = %s' % (ut.repr2(input_test_list),))
    # fnum = 1

    valid_test_list = []  # build list for printing in case of failure
    valid_test_helpstr_list = []  # for printing

    def mark_test_handled(testname):
        input_test_list.remove(testname)

    def intest(*args, **kwargs):
        helpstr = kwargs.get('help', '')
        valid_test_helpstr_list.append('   -t ' + ', '.join(args) + helpstr)
        for testname in args:
            valid_test_list.append(testname)
            ret = testname in input_test_list
            ret2 = testname in params.unknown  # Let unparsed args count towards tests
            if ret or ret2:
                if ret:
                    mark_test_handled(testname)
                else:
                    ret = ret2
                print('\n+===================')
                print(' [dev] running testname = %s' % (args,))
                print('+-------------------\n')
                return ret
        return False

    valid_test_helpstr_list.append('    # --- Simple Tests ---')

    # Explicit (simple) test functions
    if intest('export'):
        raise NotImplementedError('export')
    if intest('dbinfo'):
        dbinfo.get_dbinfo(ibs)
    if intest('headers', 'schema'):
        ibs.db.print_schema()
    if intest('info'):
        print(ibs.get_infostr())
    if intest('printcfg'):
        raise NotImplementedError('printcfg')
        # printcfg(ibs)
    if intest('tables'):
        ibs.print_tables()
    if intest('imgtbl'):
        ibs.print_image_table()

    valid_test_helpstr_list.append('    # --- Decor Tests ---')

    locals_ = locals()

    # Implicit (decorated) test functions
    for (func_aliases, func) in DEVCMD_FUNCTIONS:
        if intest(*func_aliases):
            funcname = get_funcname(func)
            # with utool.Indenter('[dev.' + funcname + ']'):
            with utool.Timer(funcname):
                # print('[dev] qid_list=%r' % (qaid_list,))
                # FIXME: , daid_list
                if len(ut.get_func_argspec(func).args) == 0:
                    ret = func()
                else:
                    ret = func(ibs, qaid_list, daid_list)
                # Add variables returned by the function to the
                # "local scope" (the exec scop)
                if hasattr(ret, 'items'):
                    for key, val in ret.items():
                        if utool.is_valid_varname(key):
                            locals_[key] = val

    valid_test_helpstr_list.append('    # --- Config Tests ---')

    # ------
    # RUNS EXPERIMENT HARNESS OVER VALID TESTNAMES SPECIFIED WITH -t
    # ------

    # Config driven test functions
    # Allow any testcfg to be in tests like: vsone_1 or vsmany_3
    test_cfg_name_list = []
    for test_cfg_name in experiment_configs.TEST_NAMES:
        if intest(test_cfg_name):
            test_cfg_name_list.append(test_cfg_name)
    # Hack to allow for very customized harness tests
    for testname in input_test_list[:]:
        if testname.startswith('custom:'):
            test_cfg_name_list.append(testname)
            mark_test_handled(testname)
    if len(test_cfg_name_list):
        fnum = pt.next_fnum()
        # Run Experiments
        # backwards compatibility yo
        acfgstr_name_list = {'OVERRIDE_HACK': (qaid_list, daid_list)}
        assert (
            False
        ), 'This way of running tests no longer works. It may be fixed in the future'
        # acfg
        harness.test_configurations(ibs, acfgstr_name_list, test_cfg_name_list)

    valid_test_helpstr_list.append('    # --- Help ---')

    if intest('help'):
        print('valid tests are:')
        print('\n'.join(valid_test_helpstr_list))
        return locals_

    if len(input_test_list) > 0:
        print('valid tests are: \n')
        print('\n'.join(valid_test_list))
        raise Exception('Unknown tests: %r ' % input_test_list)
    return locals_


# -------------------
# CUSTOM DEV FUNCS
# -------------------


# ------------------
# DEV MAIN
# ------------------


def dev_snippets(main_locals):
    """ Common variables for convineince when interacting with IPython """
    print('[dev] dev_snippets')
    species = 'zebra_grevys'
    quick = True
    fnum = 1
    # Get reference to IBEIS Controller
    ibs = main_locals['ibs']
    if 'back' in main_locals:
        # Get reference to GUI Backend
        back = main_locals['back']
        if back is not None:
            # Get reference to GUI Frontend
            front = getattr(back, 'front', None)
            ibswgt = front
            view = ibswgt.views['images']
            model = ibswgt.models['names_tree']
            selection_model = view.selectionModel()
    if ibs is not None:
        # ibs.dump_tables()
        annots = ibs.annots()
        images = ibs.images()
        aid_list = ibs.get_valid_aids()
        gid_list = ibs.get_valid_gids()
        # nid_list = ibs.get_valid_nids()
        # valid_nid_list   = ibs.get_annot_name_rowids(aid_list)
        # valid_aid_names  = ibs.get_annot_names(aid_list)
        # valid_aid_gtrues = ibs.get_annot_groundtruth(aid_list)
    return locals()


def get_sortbystr(str_list, key_list, strlbl=None, keylbl=None):
    sortx = key_list.argsort()
    ndigits = max(len(str(key_list.max())), 0 if keylbl is None else len(keylbl))
    keyfmt = '%' + str(ndigits) + 'd'
    if keylbl is not None:
        header = keylbl + ' --- ' + strlbl
    else:
        header = None

    sorted_strs = [
        (keyfmt % key + ' --- ' + str_)
        for str_, key in zip(str_list[sortx], key_list[sortx])
    ]

    def boxjoin(list_, header=None):
        topline = '+----------'
        botline = 'L__________'
        boxlines = []
        boxlines.append(topline + '\n')
        if header is not None:
            boxlines.append(header + '\n')
            boxlines.append(topline)

        body = utool.indentjoin(list_, '\n | ')
        boxlines.append(body + '\n ')
        boxlines.append(botline + '\n')
        return ''.join(boxlines)

    return boxjoin(sorted_strs, header)


# @devcmd('test_feats')
# def tst_feats(ibs, qaid_list, daid_list=None):
#     """
#     test_feats shows features using several different parameters

#     Args:
#         ibs (IBEISController):
#         qaid_list (int): query annotation id

#     CommandLine:
#         python dev.py -t test_feats --db PZ_MTEST --all --qindex 0 --show -w

#     Example:
#         >>> import wbia
#         >>> ibs = wbia.opendb('testdb1')
#         >>> qaid_list = [1]
#     """
#     from wbia import viz
#     from wbia.expt import experiment_configs
#     import utool as ut

#     NUM_PASSES = 1 if not utool.get_argflag('--show') else 2
#     varyparams_list = [experiment_configs.featparams]

#     def tst_featcfg_combo(ibs, aid, alldictcomb, count, nKpts_list, cfgstr_list):
#         for dict_ in ut.progiter(alldictcomb, lbl='FeatCFG Combo: '):
#             cfgstr_ = ut.repr2(dict_)
#             if count == 0:
#                 # On first run just record info
#                 kpts = ibs.get_annot_kpts(aid)
#                 nKpts_list.append(len(kpts))
#                 cfgstr_list.append(cfgstr_)
#             if count == 1:
#                 kpts = ibs.get_annot_kpts(aid)
#                 # If second run happens display info
#                 cfgpackstr = utool.packstr(cfgstr_, textwidth=80,
#                                               breakchars=',', newline_prefix='',
#                                               break_words=False, wordsep=',')
#                 title_suffix = (' len(kpts) = %r \n' % len(kpts)) + cfgpackstr
#                 viz.show_chip(ibs, aid, fnum=pt.next_fnum(),
#                               title_suffix=title_suffix, darken=.8,
#                               ell_linewidth=2, ell_alpha=.6, config=dict_)

#     alldictcomb = utool.flatten(map(utool.all_dict_combinations, varyparams_list))
#     for count in range(NUM_PASSES):
#         nKpts_list = []
#         cfgstr_list = []
#         for aid in qaid_list:
#             test_featcfg_combo(ibs, aid, alldictcomb, count, nKpts_list, cfgstr_list)
#             #for dict_ in alldictcomb:
#         if count == 0:
#             nKpts_list = np.array(nKpts_list)
#             cfgstr_list = np.array(cfgstr_list)
#             print(get_sortbystr(cfgstr_list, nKpts_list, 'cfg', 'nKpts'))


def run_dev(ibs):
    """
    main developer command

    CommandLine:
        python dev.py --db PZ_Master0 --controlled --print-rankhist
    """
    print('[dev] --- RUN DEV ---')
    # Get reference to controller
    if ibs is not None:
        # Get aids marked as test cases
        if ut.get_argflag('--expanded-aids'):
            ibs, qaid_list, daid_list = main_helpers.testdata_expanded_aids(ibs=ibs)
            print('[run_def] Test Annotations:')
            # print('[run_dev] * qaid_list = %s' % ut.packstr(qaid_list, 80, nlprefix='[run_dev]     '))
        else:
            qaid_list = []
            daid_list = []
        try:
            assert len(qaid_list) > 0, 'assert!'
            assert len(daid_list) > 0, 'daid_list!'
        except AssertionError as ex:
            message = ' (try using command line argument --expanded-aids to enable)'
            utool.printex(ex, 'len(qaid_list) = 0%s' % (message,), iswarning=True)
            utool.printex(ex, 'or len(daid_list) = 0%s' % (message,), iswarning=True)
            # qaid_list = ibs.get_valid_aids()[0]

        if len(qaid_list) > 0 or True:
            # Run the dev experiments
            expt_locals = run_devcmds(ibs, qaid_list, daid_list)
            # Add experiment locals to local namespace
            execstr_locals = utool.execstr_dict(expt_locals, 'expt_locals')
            exec(execstr_locals)

    return locals()


# -------------
# EXAMPLE TEXT
# -------------

EXAMPLE_TEXT = """
### DOWNLOAD A TEST DATABASE (IF REQUIRED) ###
python -m wbia.dev --t mtest
python dev.py --t nauts
./resetdbs.sh  # FIXME
python wbia/dbio/ingest_database.py  <- see module for usage

### LIST AVAIABLE DATABASES ###
python dev.py -t list_dbs

### CHOOSE A DATABASE ###
python dev.py --db PZ_Master0 --setdb
python dev.py --db GZ_ALL --setdb
python dev.py --db PZ_MTEST --setdb
python dev.py --db NAUT_test --setdb
python dev.py --db testdb1 --setdb
python dev.py --db seals2 --setdb

### DATABASE INFORMATION ###
python dev.py -t dbinfo

### EXPERIMENTS ###
python dev.py --allgt -t best
python dev.py --allgt -t vsone
python dev.py --allgt -t vsmany
python dev.py --allgt -t nsum

# Newstyle experiments
# commmand             # annot settings            # test settings
python -m wbia.dev    -a default:qaids=allgt      -t best


### COMPARE TWO CONFIGS ###
python dev.py --allgt -t nsum vsmany vsone
python dev.py --allgt -t nsum vsmany
python dev.py --allgt -t nsum vsmany vsone smk

### VARY DATABASE SIZE
python -m wbia.dev -a default:qaids=allgt,dsize=100,qper_name=1,qmin_per_name=1 -t default --db PZ_MTEST
python -m wbia.dev -a candidacy:qsize=10,dsize=100 -t default --db PZ_MTEST --verbtd


### VIZ A SET OF MATCHES ###
python dev.py --db PZ_MTEST -t query --qaid 72 110 -w
#python dev.py --allgt -t vsone vsmany
#python dev.py --allgt -t vsone --vz --vh

### RUN A SMALL AMOUNT OF VSONE TESTS ###
python dev.py --allgt -t  vsone --qindex 0:1 --vz --vh --vf --noqcache
python dev.py --allgt --qindex 0:20 --

### DUMP ANALYSIS FIGURES TO DISK ###
python dev.py --allgt -t best --vf --vz --fig-dname query_analysis_easy
python dev.py --allgt -t best --vf --vh --fig-dname query_analysis_hard
python dev.py --allgt -t best --vf --va --fig-dname query_analysis_all

python dev.py --db PZ_MTEST --set-aids-as-hard 27 28 44 49 50 51 53 54 66 72 89 97 110
python dev.py --hard -t best vsone nsum
>>>
"""

# L______________


# def run_devmain2():
#    input_test_list = ut.get_argval(('--tests', '-t',), type_=list, default=[])[:]
#    print('input_test_list = %s' % (ut.repr2(input_test_list),))
#    # fnum = 1

#    valid_test_list = []  # build list for printing in case of failure
#    valid_test_helpstr_list = []  # for printing

#    def mark_test_handled(testname):
#        input_test_list.remove(testname)

#    def intest(*args, **kwargs):
#        helpstr = kwargs.get('help', '')
#        valid_test_helpstr_list.append('   -t ' + ', '.join(args) + helpstr)
#        for testname in args:
#            valid_test_list.append(testname)
#            ret = testname in input_test_list
#            ret2 = testname in params.unknown  # Let unparsed args count towards tests
#            if ret or ret2:
#                if ret:
#                    mark_test_handled(testname)
#                else:
#                    ret = ret2
#                print('\n+===================')
#                print(' [dev2] running testname = %s' % (args,))
#                print('+-------------------\n')
#                return ret
#        return False

#    anynewhit = False
#    # Implicit (decorated) test functions
#    print('DEVCMD_FUNCTIONS2 = %r' % (DEVCMD_FUNCTIONS2,))
#    for (func_aliases, func) in DEVCMD_FUNCTIONS2:
#        if intest(*func_aliases):
#            funcname = get_funcname(func)
#            with utool.Timer(funcname):
#                if len(ut.get_func_argspec(func).args) == 0:
#                    func()
#                    anynewhit = True
#                else:
#                    func(ibs, qaid_list, daid_list)
#                    anynewhit = True
#    if anynewhit:
#        sys.exit(1)


def devmain():
    """
    The Developer Script
        A command line interface to almost everything

        -w     # wait / show the gui / figures are visible
        --cmd  # ipython shell to play with variables
        -t     # run list of tests
    """
    params.parse_args()

    print('DEVMAIN INIT - PRELOGGING')
    if params.args.logdir is not None:
        sysres.set_logdir(params.args.logdir)
    else:
        ut.start_logging(appname='wbia')
    print('DEVMAIN INIT - POSTLOGGING')

    helpstr = ut.codeblock(
        """
        Dev is meant to be run as an interactive script.

        The dev.py script runs any test you regiter with @devcmd in any combination
        of configurations specified by a Config object.

        Dev caches information in order to get quicker results.  # FIXME: Provide quicker results  # FIXME: len(line)
        """
    )

    INTRO_TITLE = 'The dev.py Script'
    # INTRO_TEXT = ''.join((ut.bubbletext(INTRO_TITLE, font='cybermedium'), helpstr))
    INTRO_TEXT = ut.bubbletext(INTRO_TITLE, font='cybermedium')

    INTRO_STR = ut.msgblock('dev.py Intro', INTRO_TEXT)

    EXAMPLE_STR = ut.msgblock('dev.py Examples', ut.codeblock(EXAMPLE_TEXT))

    if ut.NOT_QUIET:
        print(INTRO_STR)
    if ut.get_argflag(('--help', '--verbose')):
        print(EXAMPLE_STR)

    CMD = ut.get_argflag('--cmd')
    NOGUI = not ut.get_argflag('--gui')

    if len(sys.argv) == 1:
        print('Run dev.py with arguments!')
        sys.exit(1)

    # Run Precommands
    run_devprecmds()

    #
    #
    # Run IBEIS Main, create controller, and possibly gui
    print('++dev')
    main_locals = wbia.main(gui=ut.get_argflag('--gui'))
    # utool.set_process_title('wbia-dev')

    #
    #
    # Load snippet variables
    SNIPPITS = True and CMD
    if SNIPPITS:
        snippet_locals = dev_snippets(main_locals)
        snippet_execstr = utool.execstr_dict(snippet_locals, 'snippet_locals')
        exec(snippet_execstr)

    #
    #
    # Development code
    RUN_DEV = True
    if RUN_DEV:
        dev_locals = run_dev(main_locals['ibs'])
        dev_execstr = utool.execstr_dict(dev_locals, 'dev_locals')
        exec(dev_execstr)

    command = ut.get_argval('--eval', type_=str, default=None)
    if command is not None:
        result = eval(command, globals(), locals())
        print('result = %r' % (result,))
        # ibs.search_annot_notes('360')

    #
    #
    # Main Loop (IPython interaction, or some exec loop)
    # if '--nopresent' not in sys.argv or '--noshow' in sys.argv:
    ut.show_if_requested()
    if ut.get_argflag(('--show', '--wshow')):
        pt.present()
    main_execstr = wbia.main_loop(main_locals, rungui=not NOGUI, ipy=(NOGUI or CMD))
    exec(main_execstr)

    #
    #
    # Memory profile
    if ut.get_argflag('--memprof'):
        utool.print_resource_usage()
        utool.memory_profile()

    print('exiting dev')


def ggr_random_name_splits():
    """
    CommandLine:
        python -m wbia.viz.viz_graph2 ggr_random_name_splits --show

    Ignore:
        sshfs -o idmap=user lev:/ ~/lev

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.viz.viz_graph2 import *  # NOQA
        >>> ggr_random_name_splits()
    """
    import wbia.guitool as gt

    gt.ensure_qtapp()
    # nid_list = ibs.get_valid_nids(filter_empty=True)
    import wbia

    dbdir = '/media/danger/GGR/GGR-IBEIS'
    dbdir = (
        dbdir if ut.checkpath(dbdir) else ut.truepath('~/lev/media/danger/GGR/GGR-IBEIS')
    )
    ibs = wbia.opendb(dbdir=dbdir, allow_newdir=False)

    import datetime

    day1 = datetime.date(2016, 1, 30)
    day2 = datetime.date(2016, 1, 31)

    orig_filter_kw = {
        'multiple': None,
        # 'view': ['right'],
        # 'minqual': 'good',
        'is_known': True,
        'min_pername': 2,
    }
    orig_aids = ibs.filter_annots_general(
        filter_kw=ut.dict_union(
            orig_filter_kw,
            {
                'min_unixtime': ut.datetime_to_posixtime(ut.date_to_datetime(day1, 0.0)),
                'max_unixtime': ut.datetime_to_posixtime(ut.date_to_datetime(day2, 1.0)),
            },
        )
    )
    orig_all_annots = ibs.annots(orig_aids)
    orig_unique_nids, orig_grouped_annots_ = orig_all_annots.group(orig_all_annots.nids)
    # Ensure we get everything
    orig_grouped_annots = [
        ibs.annots(aids_) for aids_ in ibs.get_name_aids(orig_unique_nids)
    ]

    # pip install quantumrandom
    if False:
        import quantumrandom

        data = quantumrandom.uint16()
        seed = data.sum()
        print('seed = %r' % (seed,))
        # import Crypto.Random
        # from Crypto import Random
        # quantumrandom.get_data()
        # StrongRandom = Crypto.Random.random.StrongRandom
        # aes.reseed(3340258)
        # chars = [str(chr(x)) for x in data.view(np.uint8)]
        # aes_seed = str('').join(chars)
        # aes = Crypto.Random.Fortuna.FortunaGenerator.AESGenerator()
        # aes.reseed(aes_seed)
        # aes.pseudo_random_data(10)

    orig_rand_idxs = ut.random_indexes(len(orig_grouped_annots), seed=3340258)
    orig_sample_size = 75
    random_annot_groups = ut.take(orig_grouped_annots, orig_rand_idxs)
    orig_annot_sample = random_annot_groups[:orig_sample_size]

    # OOOPS MADE ERROR REDO ----

    filter_kw = {
        'multiple': None,
        'view': ['right'],
        'minqual': 'good',
        'is_known': True,
        'min_pername': 2,
    }
    filter_kw_ = ut.dict_union(
        filter_kw,
        {
            'min_unixtime': ut.datetime_to_posixtime(ut.date_to_datetime(day1, 0.0)),
            'max_unixtime': ut.datetime_to_posixtime(ut.date_to_datetime(day2, 1.0)),
        },
    )
    refiltered_sample = [
        ibs.filter_annots_general(annot.aids, filter_kw=filter_kw_)
        for annot in orig_annot_sample
    ]
    is_ok = np.array(ut.lmap(len, refiltered_sample)) >= 2
    ok_part_orig_sample = ut.compress(orig_annot_sample, is_ok)
    ok_part_orig_nids = [x.nids[0] for x in ok_part_orig_sample]

    # Now compute real sample
    aids = ibs.filter_annots_general(filter_kw=filter_kw_)
    all_annots = ibs.annots(aids)
    unique_nids, grouped_annots_ = all_annots.group(all_annots.nids)
    grouped_annots = grouped_annots_
    # Ensure we get everything
    # grouped_annots = [ibs.annots(aids_) for aids_ in ibs.get_name_aids(unique_nids)]

    pop = len(grouped_annots)
    pername_list = ut.lmap(len, grouped_annots)
    groups = wbia.annots.AnnotGroups(grouped_annots, ibs)
    match_tags = [ut.unique(ut.flatten(t)) for t in groups.match_tags]
    tag_case_hist = ut.dict_hist(ut.flatten(match_tags))
    print('name_pop = %r' % (pop,))
    print(
        'Annots per Multiton Name' + ut.repr3(ut.get_stats(pername_list, use_median=True))
    )
    print('Name Tag Hist ' + ut.repr3(tag_case_hist))
    print('Percent Photobomb: %.2f%%' % (tag_case_hist['photobomb'] / pop * 100))
    print('Percent Split: %.2f%%' % (tag_case_hist['splitcase'] / pop * 100))

    # Remove the ok part from this sample
    remain_unique_nids = ut.setdiff(unique_nids, ok_part_orig_nids)
    remain_grouped_annots = [
        ibs.annots(aids_) for aids_ in ibs.get_name_aids(remain_unique_nids)
    ]

    sample_size = 75
    import vtool as vt

    vt.calc_sample_from_error_bars(0.05, pop, conf_level=0.95, prior=0.05)

    remain_rand_idxs = ut.random_indexes(len(remain_grouped_annots), seed=3340258)
    remain_sample_size = sample_size - len(ok_part_orig_nids)
    remain_random_annot_groups = ut.take(remain_grouped_annots, remain_rand_idxs)
    remain_annot_sample = remain_random_annot_groups[:remain_sample_size]

    annot_sample_nofilter = ok_part_orig_sample + remain_annot_sample
    # Filter out all bad parts
    annot_sample_filter = [
        ibs.annots(ibs.filter_annots_general(annot.aids, filter_kw=filter_kw_))
        for annot in annot_sample_nofilter
    ]
    annot_sample = annot_sample_filter

    win = None
    from wbia.viz import viz_graph2

    for annots in ut.InteractiveIter(annot_sample):
        if win is not None:
            win.close()
        win = viz_graph2.make_qt_graph_interface(
            ibs, aids=annots.aids, init_mode='rereview'
        )
        print(win)

    sample_groups = wbia.annots.AnnotGroups(annot_sample, ibs)

    flat_tags = [ut.unique(ut.flatten(t)) for t in sample_groups.match_tags]

    print('Using Split and Photobomb')
    is_positive = ['photobomb' in t or 'splitcase' in t for t in flat_tags]
    num_positive = sum(is_positive)
    vt.calc_error_bars_from_sample(sample_size, num_positive, pop, conf_level=0.95)

    print('Only Photobomb')
    is_positive = ['photobomb' in t for t in flat_tags]
    num_positive = sum(is_positive)
    vt.calc_error_bars_from_sample(sample_size, num_positive, pop, conf_level=0.95)

    print('Only SplitCase')
    is_positive = ['splitcase' in t for t in flat_tags]
    num_positive = sum(is_positive)
    vt.calc_error_bars_from_sample(sample_size, num_positive, pop, conf_level=0.95)
    # gt.qtapp_loop(qwin=win)


r"""
CurrentExperiments:
    # Full best settings run
    ./dev.py -t custom --db PZ_Master0 --allgt --species=zebra_plains
    # Full best settings run without spatial verification
    ./dev.py -t custom:sv_on=False --db PZ_Master0 --allgt --species=zebra_plains

    ./dev.py -t custom --db PZ_Master0 --allgt --species=zebra_plains --hs

    # Check to see if new spatial verification helps
    ./dev.py -t custom:full_homog_checks=False custom:full_homog_checks=True --db PZ_Master0 --allgt --species=zebra_plains
    # Yay it does


    # Look for how many false negatives are in the bottom batch
    ./dev.py -t custom --db PZ_MTEST --species=zebra_plains --print-rankhist
    ./dev.py -t custom --db PZ_MTEST --controlled --print-rankhist
    ./dev.py -t custom --db PZ_Master0 --controlled --print-rankhist

    ./dev.py -t \
            custom \
            custom:rotation_invariance=True,affine_invariance=False \
            custom:rotation_invariance=True,query_rotation_heuristic=True \
            --db PZ_Master0 --controlled --print-rankhist  --print-bestcfg

    ./dev.py -t \
            custom:rotation_invariance=True,affine_invariance=False \
            custom:rotation_invariance=True,query_rotation_heuristic=True \
            --db NNP_Master3 --controlled --print-rankhist  --print-bestcfg


ElephantEarExperiments
    --show --vh
    ./dev.py -t custom:affine_invariance=True --db Elephants_drop1_ears --allgt --print-rankhist
    ./dev.py -t custom:affine_invariance=False --db Elephants_drop1_ears --allgt --print-rankhist
    ./dev.py -t custom:affine_invariance=False,histeq=True --db Elephants_drop1_ears --allgt --print-rankhist
    ./dev.py -t custom:affine_invariance=False,adapteq=True --db Elephants_drop1_ears --allgt --print-rankhist

    ./dev.py -t custom:affine_invariance=False,fg_on=False --db Elephants_drop1_ears --allgt
    ./dev.py -t custom:affine_invariance=False,histeq=True,fg_on=False --db Elephants_drop1_ears --allgt
    ./dev.py -t custom:affine_invariance=False,adapteq=True,fg_on=False --db Elephants_drop1_ears --allgt

    ./dev.py -t elph --db Elephants_drop1_ears --allgt


Sift vs Siam Experiments
    ./dev.py -t custom:feat_type=hesaff+siam128,algorithm=linear custom:feat_type=hesaff+sift --db testdb1 --allgt
    ./dev.py -t custom:feat_type=hesaff+siam128,algorithm=linear custom:feat_type=hesaff+sift --db PZ_MTEST --allgt
    ./dev.py -t custom:feat_type=hesaff+siam128,lnbnn_on=False,fg_on=False,bar_l2_on=True custom:feat_type=hesaff+sift,fg_on=False --db PZ_MTEST --allgt

    ./dev.py -t custom:feat_type=hesaff+siam128 custom:feat_type=hesaff+sift --db PZ_MTEST --allgt --print-rankhist
    ./dev.py -t custom:feat_type=hesaff+siam128 --db PZ_MTEST --allgt --print-rankhist
    ./dev.py -t custom:feat_type=hesaff+sift --db PZ_MTEST --allgt --print-rankhist

    ./dev.py -t custom:feat_type=hesaff+siam128 custom:feat_type=hesaff+sift --db PZ_Master0 --allgt

    ./dev.py -t custom:feat_type=hesaff+siam128 --db testdb1 --allgt



Without SV:
agg rank histogram = {
    (0, 1): 2276,
    (1, 5): 126,
    (5, 50): 99,
    (50, 8624): 108,
    (8624, 8625): 28,
}
With SV:
agg rank histogram = {
    (0, 1): 2300,
    (1, 5): 106,
    (5, 50): 16,
    (50, 8624): 0,
    (8624, 8625): 215,
}

Guesses:
    0 2 2 2 4 4 4 4 0 0
    0 0 4 2 2 4 4 4 2 2
    2 4 4 4 1 1 1 2 2 2
    0 0 1 1 1 2 0 0 1
"""


if __name__ == '__main__':
    """"""
    devmain()



# -*- coding: utf-8 -*-
import sys
import time
import utool as ut
import matplotlib as mpl
from wbia.plottool import custom_figure

# from .custom_constants import golden_wh


SLEEP_TIME = 0.01
__QT4_WINDOW_LIST__ = []
ut.noinject(__name__, '[fig_presenter]')


VERBOSE = ut.get_argflag(('--verbose-fig', '--verbfig', '--verb-pt'))
# (print, print_, printDBG, rrr, profile) = ut.inject(__name__, '[fig_presenter]', DEBUG=True)


def unregister_qt4_win(win):
    global __QT4_WINDOW_LIST__
    if win == 'all':
        __QT4_WINDOW_LIST__ = []
    else:
        try:
            # index = __QT4_WINDOW_LIST__.index(win)
            __QT4_WINDOW_LIST__.remove(win)
        except ValueError:
            pass


def register_qt4_win(win):
    global __QT4_WINDOW_LIST__
    __QT4_WINDOW_LIST__.append(win)


# ---- GENERAL FIGURE COMMANDS ----


def set_geometry(fnum, x, y, w, h):
    fig = custom_figure.ensure_fig(fnum)
    qtwin = get_figure_window(fig)
    qtwin.setGeometry(x, y, w, h)


def get_geometry(fnum):
    fig = custom_figure.ensure_fig(fnum)
    qtwin = get_figure_window(fig)
    (x1, y1, x2, y2) = qtwin.geometry().getCoords()
    (x, y, w, h) = (x1, y1, x2 - x1, y2 - y1)
    return (x, y, w, h)


# def get_screen_info():
#    # TODO Move dependency to guitool
#    desktop = QtWidgets.QDesktopWidget()
#    mask = desktop.mask()  # NOQA
#    layout_direction = desktop.layoutDirection()  # NOQA
#    screen_number = desktop.screenNumber()  # NOQA
#    normal_geometry = desktop.normalGeometry()  # NOQA
#    num_screens = desktop.screenCount()  # NOQA
#    avail_rect = desktop.availableGeometry()  # NOQA
#    screen_rect = desktop.screenGeometry()  # NOQA
#    QtWidgets.QDesktopWidget().availableGeometry().center()  # NOQA
#    normal_geometry = desktop.normalGeometry()  # NOQA


# @profile
def get_all_figures():
    manager_list = mpl._pylab_helpers.Gcf.get_all_fig_managers()
    all_figures = []
    # Make sure you dont show figures that this module closed
    for manager in manager_list:
        try:
            fig = manager.canvas.figure
        except AttributeError:
            continue
        if not fig.__dict__.get('df2_closed', False):
            all_figures.append(fig)
    # Return all the figures sorted by their number
    all_figures = sorted(all_figures, key=lambda fig: fig.number)
    return all_figures


def get_all_qt4_wins():
    return __QT4_WINDOW_LIST__


def all_figures_show():
    if VERBOSE:
        print('all_figures_show')
    if not ut.get_argflag('--noshow'):
        for fig in get_all_figures():
            time.sleep(SLEEP_TIME)
            show_figure(fig)
            # fig.show()
            # fig.canvas.draw()


def show_figure(fig):
    try:
        fig.show()
        fig.canvas.draw()
    except AttributeError as ex:
        if not hasattr(fig, '_no_raise_plottool'):
            ut.printex(
                ex, '[pt] probably registered made figure with Qt.', iswarning=True
            )


def all_figures_tight_layout():
    if '--noshow' not in sys.argv:
        for fig in iter(get_all_figures()):
            fig.tight_layout()
            time.sleep(SLEEP_TIME)


def get_main_win_base():
    if hasattr(mpl.backends, 'backend_qt4'):
        backend = mpl.backends.backend_qt4
    else:
        backend = mpl.backends.backend_qt5
    try:
        QMainWin = backend.MainWindow
    except Exception as ex:
        try:
            ut.printex(ex, 'warning', '[fig_presenter]')
            # from wbia.guitool.__PYQT__ import QtGui
            QMainWin = backend.QtWidgets.QMainWindow
        except Exception as ex1:
            ut.printex(ex1, 'warning', '[fig_presenter]')
            QMainWin = object
    return QMainWin


def get_all_windows():
    """ Returns all mpl figures and registered qt windows """
    try:
        all_figures = get_all_figures()
        all_qt4wins = get_all_qt4_wins()
        all_wins = all_qt4wins + [get_figure_window(fig) for fig in all_figures]
        return all_wins
    except AttributeError as ex:
        ut.printex(ex, 'probably using a windowless backend', iswarning=True)
        return []


# @profile
def all_figures_tile(
    max_rows=None,
    row_first=True,
    no_tile=False,
    monitor_num=None,
    percent_w=None,
    percent_h=None,
    hide_toolbar=True,
):
    """
    Lays out all figures in a grid. if wh is a scalar, a golden ratio is used
    """
    # print('[plottool] all_figures_tile()')
    if no_tile:
        return

    current_backend = mpl.get_backend()
    if not current_backend.startswith('Qt'):
        # print('current_backend=%r is not a Qt backend. cannot tile.' % current_backend)
        return

    all_wins = get_all_windows()
    num_wins = len(all_wins)
    if num_wins == 0:
        return

    from wbia.plottool import screeninfo

    valid_positions = screeninfo.get_valid_fig_positions(
        num_wins,
        max_rows,
        row_first,
        monitor_num,
        percent_w=percent_w,
        percent_h=percent_h,
    )

    QMainWin = get_main_win_base()
    for ix, win in enumerate(all_wins):
        isqt4_mpl = isinstance(win, QMainWin)
        from wbia.guitool.__PYQT__ import QtGui  # NOQA
        from wbia.guitool.__PYQT__ import QtWidgets  # NOQA

        isqt4_back = isinstance(win, QtWidgets.QMainWindow)
        isqt4_widget = isinstance(win, QtWidgets.QWidget)
        (x, y, w, h) = valid_positions[ix]
        # printDBG('tile %d-th win: xywh=%r' % (ix, (x, y, w, h)))
        if not isqt4_mpl and not isqt4_back and not isqt4_widget:
            raise NotImplementedError('%r-th Backend %r is not a Qt Window' % (ix, win))
        try:
            if hide_toolbar:
                toolbar = win.findChild(QtWidgets.QToolBar)
                toolbar.setVisible(False)
            win.setGeometry(x, y, w, h)
        except Exception as ex:
            ut.printex(ex)


def all_figures_bring_to_front():
    try:
        all_figures = get_all_figures()
        for fig in iter(all_figures):
            bring_to_front(fig)
    except Exception as ex:
        if not hasattr(fig, '_no_raise_plottool'):
            ut.printex(ex, iswarning=True)


def close_all_figures():
    print('[pt] close_all_figures')
    all_figures = get_all_figures()
    for fig in iter(all_figures):
        close_figure(fig)


def close_figure(fig):
    print('[pt] close_figure')
    fig.clf()
    fig.df2_closed = True
    qtwin = get_figure_window(fig)
    qtwin.close()


def get_figure_window(fig):
    try:
        qwin = fig.canvas.manager.window
    except AttributeError:
        qwin = fig.canvas.window()
    return qwin


def bring_to_front(fig):
    if VERBOSE:
        print('[pt] bring_to_front')
    # what is difference between show and show normal?
    qtwin = get_figure_window(fig)
    qtwin.raise_()
    # if not ut.WIN32:
    # NOT sure on the correct order of these
    # can cause the figure geometry to be unset
    from wbia.guitool.__PYQT__.QtCore import Qt

    qtwin.activateWindow()
    qtwin.setWindowFlags(Qt.WindowStaysOnTopHint)
    qtwin.setWindowFlags(Qt.WindowFlags(0))
    qtwin.show()


def show():
    if VERBOSE:
        print('[pt] show')
    all_figures_show()
    all_figures_bring_to_front()
    # plt.show()


def reset():
    if VERBOSE:
        print('[pt] reset')
    close_all_figures()


def draw():
    if VERBOSE:
        print('[pt] draw')
    all_figures_show()


def update():
    if VERBOSE:
        print('[pt] update')
    draw()
    all_figures_bring_to_front()


def iupdate():
    if VERBOSE:
        print('[pt] iupdate')
    if ut.inIPython():
        update()


iup = iupdate


def present(*args, **kwargs):
    """
    basically calls show if not embeded.

    Kwargs:
        max_rows, row_first, no_tile, monitor_num, percent_w, percent_h,
        hide_toolbar

    CommandLine:
        python -m wbia.plottool.fig_presenter present

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.plottool.fig_presenter import *  # NOQA
        >>> result = present()
        >>> print(result)
        >>> import wbia.plottool as pt
        >>> pt.show_if_requested()
    """
    if VERBOSE:
        print('[pt] present')
    if not ut.get_argflag('--noshow'):
        # print('[fig_presenter] Presenting figures...')
        # with warnings.catch_warnings():
        #    warnings.simplefilter("ignore")
        all_figures_tile(*args, **kwargs)
        # Both of these lines cause the weird non-refresh black border behavior
        all_figures_show()
        all_figures_bring_to_front()



# -*- coding: utf-8 -*-
"""
helpers for painting on top of images for groundtruthing

References:
    http://stackoverflow.com/questions/22232812/drawing-on-image-with-matplotlib-and-opencv-update-image
    http://stackoverflow.com/questions/34933254/force-matplotlib-to-block-in-a-pyqt-thread-process
    http://matplotlib.org/examples/user_interfaces/embedding_in_qt4.html
    http://stackoverflow.com/questions/22410663/block-qmainwindow-while-child-widget-is-alive-pyqt
    http://stackoverflow.com/questions/20289939/pause-execution-until-button-press
"""
import utool as ut
import matplotlib.pyplot as plt
import numpy as np

try:
    import vtool as vt
except ImportError:
    pass
from wbia.plottool import abstract_interaction
import math
from six.moves import range, zip, input  # NOQA

ut.noinject('impaint')


PAINTER_BASE = abstract_interaction.AbstractInteraction


class PaintInteraction(PAINTER_BASE):
    """
    References:
        http://stackoverflow.com/questions/22232812/drawing-on-image-with-mpl

    CommandLine:
        python -m wbia.plottool.interact_impaint --exec-draw_demo --show
    """

    def __init__(self, img, **kwargs):
        super(PaintInteraction, self).__init__(**kwargs)
        init_mask = kwargs.get('init_mask', None)
        if init_mask is None:
            mask = np.full(img.shape, 255, dtype=np.uint8)
        else:
            mask = init_mask
        self.mask = mask
        self.img = img
        self.brush_size = 75
        import wbia.plottool as pt

        self.valid_colors1 = ut.odict(
            [
                # ('background', (255 * pt.BLACK).tolist()),
                ('scenery', (255 * pt.BLACK).tolist()),
                ('photobomb', (255 * pt.RED).tolist()),
            ]
        )
        self.valid_colors2 = ut.odict([('foreground', (255 * pt.WHITE).tolist())])
        self.color1_idx = 0
        self.color1 = self.valid_colors1['scenery']
        self.color2 = self.valid_colors2['foreground']
        self.background = None
        self.last_stroke = None
        self.finished_callback = None
        self._imshow_running = True

    def update_title(self):
        import wbia.plottool as pt

        key = (self.valid_colors1.keys())[self.color1_idx]
        pt.plt.title(
            'Click on the image to draw. exit to finish.\n'
            'Right click erases, scroll wheel resizes.'
            't changes current_color=%r' % (key,)
        )

    def static_plot(self, fnum=None, pnum=(1, 1, 1)):
        import wbia.plottool as pt

        self.ax = pt.gca()
        # self.ax.imshow(img, interpolation='nearest', alpha=1)
        # self.ax.imshow(mask, interpolation='nearest', alpha=0.6)
        pt.imshow(self.img, ax=self.ax, interpolation='nearest', alpha=1)
        pt.imshow(self.mask, ax=self.ax, interpolation='nearest', alpha=0.6)
        self.update_title()
        self.ax.grid(False)

    def update_image(self):
        import wbia.plottool as pt

        # print('update_image')
        self.ax.images.pop()
        # self.ax.imshow(self.mask, interpolation='nearest', alpha=0.6)
        pt.imshow(self.mask, ax=self.ax, interpolation='nearest', alpha=0.6)
        self.draw()
        # self.do_blit()
        # self.update()
        # self.ax.imshow(vt.blend_images_multiply(self.img, self.mask))
        # self.ax.grid(False)
        # self.ax.set_xticks([])
        # self.ax.set_yticks([])

    def on_close(self, event=None):
        if self.finished_callback is not None:
            self.finished_callback(self.mask)
        super(PaintInteraction, self).on_close(event)

    def do_blit(self):
        if self.debug > 3:
            print('[pt.impaint] do_blit')
        if self.background is None:
            self.background = self.fig.canvas.copy_from_bbox(self.ax.bbox)
        else:
            self.fig.canvas.restore_region(self.background)
            pass
        self.fig.canvas.blit(self.ax.bbox)

    def on_draw(self, event):
        # print('on draw')
        self.background = self.fig.canvas.copy_from_bbox(self.ax.bbox)

    def apply_stroke(self, x, y, color):
        import cv2

        if self.debug > 3:
            print('[pt.impaint] apply stroke')
        center = (x, y)
        radius = int(self.brush_size / 2)
        thickness = -1
        color_bgr = color[0:3][::-1]
        cv2.circle(self.mask, center, radius, color_bgr, thickness)
        if self.last_stroke is not None:
            if self.last_stroke[0] == color:
                old_center = self.last_stroke[1]
                line_thickness = int(self.brush_size)
                cv2.line(self.mask, center, old_center, color_bgr, line_thickness)
        self.last_stroke = (color, (x, y))

    def on_click_inside(self, event, ax):
        x = int(math.floor(event.xdata))
        y = int(math.floor(event.ydata))
        if event.button == self.LEFT_BUTTON:
            self.apply_stroke(x, y, self.color1)
        if event.button == self.RIGHT_BUTTON:
            self.apply_stroke(x, y, self.color2)
        self.update_image()
        # self.draw()
        # self.print_status()

    def on_scroll(self, event):
        self.brush_size = max(self.brush_size + event.step, 1)
        print('self.brush_size = %r' % (self.brush_size,))

    def on_key_press(self, event):
        if event.key == 't':
            print('toggle color')
            self.color1_idx = (self.color1_idx + 1) % len(self.valid_colors1)
            key = (self.valid_colors1.keys())[self.color1_idx]
            self.color1 = self.valid_colors1[key]
            print('self.color1_idx = %r' % (self.color1_idx,))
            print('key = %r' % (key,))
            self.update_title()
            self.draw()

    def on_drag_stop(self, event):
        self.last_stroke = None

    def on_drag_inside(self, event):
        # self.print_status()
        x = int(math.floor(event.xdata))
        y = int(math.floor(event.ydata))
        if event.button == self.LEFT_BUTTON:
            self.apply_stroke(x, y, self.color1)
        elif event.button == self.RIGHT_BUTTON:
            self.apply_stroke(x, y, self.color2)
        self.update_image()
        # self.do_blit()
        # self.draw()


def impaint_mask2(img, init_mask=None):
    """
    python -m wbia.plottool.interact_impaint --exec-draw_demo --show
    """
    if False:
        QT = False  # NOQA
        # if QT:
        #    from wbia.guitool import mpl_embed
        #    import wbia.guitool
        #    guitool.ensure_qapp()  # must be ensured before any embeding
        #    wgt = mpl_embed.QtAbstractMplInteraction()
        #    fig = wgt.fig
        #    ax = wgt.axes
        # else:
        #    fig = plt.figure(1)
        #    ax = plt.subplot(111)
        # if init_mask is None:
        #    mask = np.zeros(img.shape, np.uint8) + 255
        # else:
        #    mask = init_mask
        # ax.imshow(img, interpolation='nearest', alpha=1)
        # ax.imshow(mask, interpolation='nearest', alpha=0.6)
        # ax.grid(False)
        # ax.set_xticks([])
        # ax.set_yticks([])

        # pstartntr = _OldPainter(fig, ax, mask)
        # ax.set_title('Click on the image to draw. exit to finish')
        # print('Starting interaction')
        # if not QT:
        #    plt.show(block=True)
        # else:
        #    guitool.qtapp_loop(wgt, frequency=100, init_signals=True)
        #    wgt.show()
        # # input('hack to block... press enter when done')
    else:
        pntr = PaintInteraction(img, init_mask=init_mask)
        # pntr.show_page()
        # print('Starting interaction')
        pntr.start()
        pntr.show()

        # Hacky code to block until the interaction is actually done
        # pntr.show()
        import time
        from wbia.guitool.__PYQT__ import QtGui as QtWidgets

        while pntr.is_running:
            QtWidgets.qApp.processEvents()
            time.sleep(0.05)
        # plt.show()
    print('Finished interaction')
    return pntr.mask


def draw_demo():
    r"""
    CommandLine:
        python -m wbia.plottool.interact_impaint --exec-draw_demo --show

    Example:
        >>> # SCRIPT
        >>> from wbia.plottool.interact_impaint import *  # NOQA
        >>> result = draw_demo()
        >>> print(result)
        >>> import wbia.plottool as pt
        >>> pt.show_if_requested()
    """
    fpath = ut.grab_test_imgpath('zebra.png')
    img = vt.imread(fpath)
    mask = impaint_mask2(img)
    print('mask = %r' % (mask,))
    print('mask.sum() = %r' % (mask.sum(),))
    if False:
        plt.imshow(vt.blend_images_multiply(img, mask))
        ax = plt.gca()
        ax.grid(False)
        ax.set_xticks([])
        ax.set_yticks([])



# -*- coding: utf-8 -*-
import logging
import utool as ut

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')


def reasign_names1(ibs, aid_list=None, old_img2_names=None, common_prefix=''):
    r"""
    Changes the names in the IA-database to correspond to an older
    naming convention.  If splits and merges were preformed tries to
    find the maximally consistent renaming scheme.

    Notes:
        For each annotation:
        * get the image
        * get the image full path
        * strip the full path down to the file name prefix:
             [ example /foo/bar/pic.jpg -> pic ]
        * make the name of the individual associated with that annotation be the
          file name prefix
        * save the new names to the image analysis database
        * wildbook will make a request to get all of the annotations, image
          file names, image names and animal ids

    CommandLine:
        python -m wbia.scripts.name_recitifer rectify_names --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.scripts.name_recitifer import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> aid_list = None
        >>> common_prefix = ''
        >>> old_img2_names = None #['img_fred.png', ']
        >>> result = reasign_names1(ibs, aid_list, img_list, name_list)
    """
    if aid_list is None:
        aid_list = ibs.get_valid_aids()
    # Group annotations by their current IA-name
    nid_list = ibs.get_annot_name_rowids(aid_list)
    nid2_aids = ut.group_items(aid_list, nid_list)
    unique_nids = list(nid2_aids.keys())
    grouped_aids = list(nid2_aids.values())

    # Get grouped images
    grouped_imgnames = ibs.unflat_map(ibs.get_annot_image_names, grouped_aids)

    # Assume a mapping from old image names to old names is given.
    # Or just hack it in the Lewa case.
    if old_img2_names is None:

        def get_name_from_gname(gname):
            from os.path import splitext

            gname_, ext = splitext(gname)
            assert gname_.startswith(common_prefix), 'prefix assumption is invalidated'
            gname_ = gname_[len(common_prefix) :]
            return gname_

        # Create mapping from image name to the desired "name" for the image.
        old_img2_names = {
            gname: get_name_from_gname(gname) for gname in ut.flatten(grouped_imgnames)
        }

    # Make the name of the individual associated with that annotation be the
    # file name prefix
    grouped_oldnames = [ut.take(old_img2_names, gnames) for gnames in grouped_imgnames]

    # The task is now to map each name in unique_nids to one of these names
    # subject to the contraint that each name can only be used once.  This is
    # solved using a maximum bipartite matching. The new names are the left
    # nodes, the old name are the right nodes, and grouped_oldnames definse the
    # adjacency matrix.
    # NOTE: In rare cases it may be impossible to find a correct labeling using
    # only old names.  In this case new names will be created.
    new_name_text = find_consistent_labeling(grouped_oldnames)

    dry = False
    if not dry:
        # Save the new names to the image analysis database
        ibs.set_name_texts(unique_nids, new_name_text)


def reasign_names2(ibs, gname_name_pairs, aid_list=None):
    """

    Notes:
        * Given a list of pairs:  image file names (full path), animal name.
        * Go through all the images in the database and create a dictionary
        that associates the file name (full path) of the image in the database
        with the
          annotation or annotations associated with that image.
        * Go through the list of pairs:
          For each image file name, look up in the dictionary the image file
          name and assign the annotation associated with the image file name
          the animal name
        * Throughout this, keep a list of annotations that have been changed
        * Wildbook will issue a pull request to get these annotation.

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.scripts.name_recitifer import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> aid_list = None
        >>> common_prefix = ''
        >>> gname_name_pairs = [
        >>>     ('easy1.JPG', 'easy'),
        >>>     ('easy2.JPG', 'easy'),
        >>>     ('easy3.JPG', 'easy'),
        >>>     ('hard1.JPG', 'hard')
        >>> ]
        >>> changed_pairs = reasign_names2(gname_name_pairs)
    """
    from os.path import basename

    if aid_list is None:
        aid_list = ibs.get_valid_aids()
    annot_gnames = ibs.get_annot_image_names(aid_list)
    # Other image name getters that may be useful
    # ibs.get_annot_image_paths(aid_list)
    # ibs.get_image_uris_original(ibs.get_annot_gids(aid_list))
    gname2_aids = ut.group_items(aid_list, annot_gnames)

    changed_aids = []
    changed_names = []

    for gname, name in gname_name_pairs:
        # make sure its just the last part of the name.
        # Ignore preceding path
        gname = basename(gname)
        aids = gname2_aids[gname]
        texts = ibs.get_annot_name_texts(aids)
        flags = [text != name for text in texts]
        aids_ = ut.compress(aids, flags)
        if len(aids_):
            changed_aids.extend(aids_)
            changed_names.extend([name] * len(aids_))

    dry = False
    if not dry:
        # Save the new names to the image analysis database
        ibs.set_annot_name_texts(changed_aids, changed_names)

    # Returned list tells you who was changed.
    changed_pairs = list(zip(changed_names, changed_aids))
    return changed_pairs


def testdata_oldnames(
    n_incon_groups=10,
    n_con_groups=2,
    n_per_con=5,
    n_per_incon=5,
    con_sep=4,
    n_empty_groups=0,
):
    import numpy as np

    rng = np.random.RandomState(42)

    rng.randint(1, con_sep + 1)

    n_incon_labels = rng.randint(0, n_incon_groups + 1)
    incon_labels = list(range(n_incon_labels))

    # Build up inconsistent groups that may share labels with other groups
    n_per_incon_list = [
        rng.randint(min(2, n_per_incon), n_per_incon + 1) for _ in range(n_incon_groups)
    ]
    incon_groups = [
        rng.choice(incon_labels, n, replace=True).tolist() for n in n_per_incon_list
    ]

    # Build up consistent groups that may have multiple lables, but does not
    # share labels with any other group
    con_groups = []
    offset = n_incon_labels + 1
    for _ in range(n_con_groups):
        this_n_per = rng.randint(1, n_per_con + 1)
        this_n_avail = rng.randint(1, con_sep + 1)
        this_avail_labels = list(range(offset, offset + this_n_avail))
        this_labels = rng.choice(this_avail_labels, this_n_per, replace=True)
        con_groups.append(this_labels.tolist())
        offset += this_n_avail

    empty_groups = [[] for _ in range(n_empty_groups)]

    grouped_oldnames = incon_groups + con_groups + empty_groups
    # rng.shuffle(grouped_oldnames)
    return grouped_oldnames


def simple_munkres(part_oldnames):
    """
    Defines a munkres problem to solve name rectification.

    Notes:
        We create a matrix where each rows represents a group of annotations in
        the same PCC and each column represents an original name. If there are
        more PCCs than original names the columns are padded with extra values.
        The matrix is first initialized to be negative infinity representing
        impossible assignments. Then for each column representing a padded
        name, we set we its value to $1$ indicating that each new name could be
        assigned to a padded name for some small profit.  Finally, let $f_{rc}$
        be the the number of annotations in row $r$ with an original name of
        $c$. Each matrix value $(r, c)$ is set to $f_{rc} + 1$ if $f_{rc} > 0$,
        to represent how much each name ``wants'' to be labeled with a
        particular original name, and the extra one ensures that these original
        names are always preferred over padded names.

    CommandLine:
        python -m wbia.scripts.name_recitifer simple_munkres

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.scripts.name_recitifer import *  # NOQA
        >>> part_oldnames = [['a', 'b'], ['b', 'c'], ['c', 'a', 'a']]
        >>> new_names = simple_munkres(part_oldnames)
        >>> result = ut.repr2(new_names)
        >>> print(new_names)
        ['b', 'c', 'a']

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.scripts.name_recitifer import *  # NOQA
        >>> part_oldnames = [[], ['a', 'a'], [],
        >>>                  ['a', 'a', 'a', 'a', 'a', 'a', 'a', 'b'], ['a']]
        >>> new_names = simple_munkres(part_oldnames)
        >>> result = ut.repr2(new_names)
        >>> print(new_names)
        [None, 'a', None, 'b', None]

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.scripts.name_recitifer import *  # NOQA
        >>> part_oldnames = [[], ['b'], ['a', 'b', 'c'], ['b', 'c'], ['c', 'e', 'e']]
        >>> new_names = find_consistent_labeling(part_oldnames)
        >>> result = ut.repr2(new_names)
        >>> print(new_names)
        ['_extra_name0', 'b', 'a', 'c', 'e']

        Profit Matrix
            b   a   c   e  _0
        0 -10 -10 -10 -10   1
        1   2 -10 -10 -10   1
        2   2   2   2 -10   1
        3   2 -10   2 -10   1
        4 -10 -10   2   3   1
    """
    import numpy as np
    import scipy.optimize

    unique_old_names = ut.unique(ut.flatten(part_oldnames))
    num_new_names = len(part_oldnames)
    num_old_names = len(unique_old_names)

    # Create padded dummy values.  This accounts for the case where it is
    # impossible to uniquely map to the old db
    num_pad = max(num_new_names - num_old_names, 0)
    total = num_old_names + num_pad
    shape = (total, total)

    # Allocate assignment matrix.
    # rows are new-names and cols are old-names.
    # Initially the profit of any assignment is effectively -inf
    # This effectively marks all assignments as invalid
    profit_matrix = np.full(shape, -2 * total, dtype=np.int)
    # Overwrite valid assignments with positive profits
    oldname2_idx = ut.make_index_lookup(unique_old_names)
    name_freq_list = [ut.dict_hist(names) for names in part_oldnames]
    # Initialize profit of a valid assignment as 1 + freq
    # This incentivizes using a previously used name
    for rowx, name_freq in enumerate(name_freq_list):
        for name, freq in name_freq.items():
            colx = oldname2_idx[name]
            profit_matrix[rowx, colx] = freq + 1
    # Set a much smaller profit for using an extra name
    # This allows the solution to always exist
    profit_matrix[:, num_old_names:total] = 1

    # Convert to minimization problem
    big_value = (profit_matrix.max()) - (profit_matrix.min())
    cost_matrix = big_value - profit_matrix

    # Use scipy implementation of munkres algorithm.
    rx2_cx = dict(zip(*scipy.optimize.linear_sum_assignment(cost_matrix)))

    # Each row (new-name) has now been assigned a column (old-name)
    # Map this back to the input-space (using None to indicate extras)
    cx2_name = dict(enumerate(unique_old_names))

    if False:
        import pandas as pd

        columns = unique_old_names + ['_%r' % x for x in range(num_pad)]
        logger.info('Profit Matrix')
        logger.info(pd.DataFrame(profit_matrix, columns=columns))

        logger.info('Cost Matrix')
        logger.info(pd.DataFrame(cost_matrix, columns=columns))

    assignment_ = [cx2_name.get(rx2_cx[rx], None) for rx in range(num_new_names)]
    return assignment_


def find_consistent_labeling(grouped_oldnames, extra_prefix='_extra_name', verbose=False):
    r"""
    Solves a a maximum bipirtite matching problem to find a consistent
    name assignment that minimizes the number of annotations with different
    names. For each new grouping of annotations we assign

    For each group of annotations we must assign them all the same name, either from




    To reduce the running time

    Args:
        gropued_oldnames (list): A group of old names where the grouping is
            based on new names. For instance:

                Given:
                    aids      = [1, 2, 3, 4, 5]
                    old_names = [0, 1, 1, 1, 0]
                    new_names = [0, 0, 1, 1, 0]

                The grouping is
                    [[0, 1, 0], [1, 1]]

                This lets us keep the old names in a split case and
                re-use exising names and make minimal changes to
                current annotation names while still being consistent
                with the new and improved grouping.

                The output will be:
                    [0, 1]

                Meaning that all annots in the first group are assigned the
                name 0 and all annots in the second group are assigned the name
                1.

    References:
        http://stackoverflow.com/questions/1398822/assignment-problem-numpy

    CommandLine:
        python -m wbia.scripts.name_recitifer find_consistent_labeling


    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.scripts.name_recitifer import *  # NOQA
        >>> grouped_oldnames = testdata_oldnames(25, 15,  5, n_per_incon=5)
        >>> new_names = find_consistent_labeling(grouped_oldnames, verbose=1)
        >>> grouped_oldnames = testdata_oldnames(0, 15,  5, n_per_incon=1)
        >>> new_names = find_consistent_labeling(grouped_oldnames, verbose=1)
        >>> grouped_oldnames = testdata_oldnames(0, 0, 0, n_per_incon=1)
        >>> new_names = find_consistent_labeling(grouped_oldnames, verbose=1)

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.scripts.name_recitifer import *  # NOQA
        >>> ydata = []
        >>> xdata = list(range(10, 150, 50))
        >>> for x in xdata:
        >>>     print('x = %r' % (x,))
        >>>     grouped_oldnames = testdata_oldnames(x, 15,  5, n_per_incon=5)
        >>>     t = ut.Timerit(3, verbose=1)
        >>>     for timer in t:
        >>>         with timer:
        >>>             new_names = find_consistent_labeling(grouped_oldnames)
        >>>     ydata.append(t.ave_secs)
        >>> ut.quit_if_noshow()
        >>> import wbia.plottool as pt
        >>> pt.qtensure()
        >>> pt.multi_plot(xdata, [ydata])
        >>> ut.show_if_requested()

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.scripts.name_recitifer import *  # NOQA
        >>> grouped_oldnames = [['a', 'b', 'c'], ['b', 'c'], ['c', 'e', 'e']]
        >>> new_names = find_consistent_labeling(grouped_oldnames, verbose=1)
        >>> result = ut.repr2(new_names)
        >>> print(new_names)
        ['a', 'b', 'e']

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.scripts.name_recitifer import *  # NOQA
        >>> grouped_oldnames = [['a', 'b'], ['a', 'a', 'b'], ['a']]
        >>> new_names = find_consistent_labeling(grouped_oldnames)
        >>> result = ut.repr2(new_names)
        >>> print(new_names)
        ['b', 'a', '_extra_name0']

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.scripts.name_recitifer import *  # NOQA
        >>> grouped_oldnames = [['a', 'b'], ['e'], ['a', 'a', 'b'], [], ['a'], ['d']]
        >>> new_names = find_consistent_labeling(grouped_oldnames)
        >>> result = ut.repr2(new_names)
        >>> print(new_names)
        ['b', 'e', 'a', '_extra_name0', '_extra_name1', 'd']

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.scripts.name_recitifer import *  # NOQA
        >>> grouped_oldnames = [[], ['a', 'a'], [],
        >>>                     ['a', 'a', 'a', 'a', 'a', 'a', 'a', 'b'], ['a']]
        >>> new_names = find_consistent_labeling(grouped_oldnames)
        >>> result = ut.repr2(new_names)
        >>> print(new_names)
        ['_extra_name0', 'a', '_extra_name1', 'b', '_extra_name2']
    """
    unique_old_names = ut.unique(ut.flatten(grouped_oldnames))
    n_old_names = len(unique_old_names)
    n_new_names = len(grouped_oldnames)

    # Initialize assignment to all Nones
    assignment = [None for _ in range(n_new_names)]

    if verbose:
        logger.info('finding maximally consistent labeling')
        logger.info('n_old_names = %r' % (n_old_names,))
        logger.info('n_new_names = %r' % (n_new_names,))

    # For each old_name, determine now many new_names use it.
    oldname_sets = list(map(set, grouped_oldnames))
    oldname_usage = ut.dict_hist(ut.flatten(oldname_sets))

    # Any name used more than once is a conflict and must be resolved
    conflict_oldnames = {k for k, v in oldname_usage.items() if v > 1}

    # Partition into trivial and non-trivial cases
    nontrivial_oldnames = []
    nontrivial_new_idxs = []

    trivial_oldnames = []
    trivial_new_idxs = []
    for new_idx, group in enumerate(grouped_oldnames):
        if set(group).intersection(conflict_oldnames):
            nontrivial_oldnames.append(group)
            nontrivial_new_idxs.append(new_idx)
        else:
            trivial_oldnames.append(group)
            trivial_new_idxs.append(new_idx)

    # Rectify trivial cases
    # Any new-name that does not share any of its old-names with other
    # new-names can be resolved trivially
    n_trivial_unchanged = 0
    n_trivial_ignored = 0
    n_trivial_merges = 0
    for group, new_idx in zip(trivial_oldnames, trivial_new_idxs):
        if len(group) > 0:
            # new-names that use more than one old-name are simple merges
            h = ut.dict_hist(group)
            if len(h) > 1:
                n_trivial_merges += 1
            else:
                n_trivial_unchanged += 1
            hitems = list(h.items())
            hvals = [i[1] for i in hitems]
            maxval = max(hvals)
            g = min([k for k, v in hitems if v == maxval])
            assignment[new_idx] = g
        else:
            # new-names that use no old-names can be ignored
            n_trivial_ignored += 1

    if verbose:
        n_trivial = len(trivial_oldnames)
        n_nontrivial = len(nontrivial_oldnames)
        logger.info('rectify %d trivial groups' % (n_trivial,))
        logger.info('  * n_trivial_unchanged = %r' % (n_trivial_unchanged,))
        logger.info('  * n_trivial_merges = %r' % (n_trivial_merges,))
        logger.info('  * n_trivial_ignored = %r' % (n_trivial_ignored,))
        logger.info('rectify %d non-trivial groups' % (n_nontrivial,))

    # Partition nontrivial_oldnames into smaller disjoint sets
    nontrivial_oldnames_sets = list(map(set, nontrivial_oldnames))
    import networkx as nx

    g = nx.Graph()
    g.add_nodes_from(range(len(nontrivial_oldnames_sets)))
    for u, group1 in enumerate(nontrivial_oldnames_sets):
        rest = nontrivial_oldnames_sets[u + 1 :]
        for v, group2 in enumerate(rest, start=u + 1):
            if group1.intersection(group2):
                g.add_edge(u, v)
    nontrivial_partition = list(nx.connected_components(g))
    if verbose:
        logger.info(
            '  * partitioned non-trivial into %d subgroups' % (len(nontrivial_partition))
        )
        part_size_stats = ut.get_stats(map(len, nontrivial_partition))
        stats_str = ut.repr2(part_size_stats, precision=2, strkeys=True)
        logger.info('  * partition size stats = %s' % (stats_str,))

    # Rectify nontrivial cases
    for part_idxs in ut.ProgIter(
        nontrivial_partition, labels='rectify parts', enabled=verbose
    ):
        part_oldnames = ut.take(nontrivial_oldnames, part_idxs)
        part_newidxs = ut.take(nontrivial_new_idxs, part_idxs)
        # Rectify this part
        assignment_ = simple_munkres(part_oldnames)
        for new_idx, new_name in zip(part_newidxs, assignment_):
            assignment[new_idx] = new_name

    # Any unassigned name is now given a new unique label with a prefix
    if extra_prefix is not None:
        num_extra = 0
        for idx, val in enumerate(assignment):
            if val is None:
                assignment[idx] = '%s%d' % (extra_prefix, num_extra)
                num_extra += 1
    return assignment


def find_consistent_labeling_old(
    grouped_oldnames, extra_prefix='_extra_name', verbose=False
):
    import numpy as np
    import scipy.optimize

    unique_old_names = ut.unique(ut.flatten(grouped_oldnames))

    # TODO: find names that are only used once, and just ignore those for
    # optimization.
    # unique_set = set(unique_old_names)
    oldname_sets = list(map(set, grouped_oldnames))
    usage_hist = ut.dict_hist(ut.flatten(oldname_sets))
    conflicts = {k for k, v in usage_hist.items() if v > 1}
    # nonconflicts = {k for k, v in usage_hist.items() if v == 1}

    conflict_groups = []
    orig_idxs = []
    assignment = [None] * len(grouped_oldnames)
    ntrivial = 0
    for idx, group in enumerate(grouped_oldnames):
        if set(group).intersection(conflicts):
            orig_idxs.append(idx)
            conflict_groups.append(group)
        else:
            ntrivial += 1
            if len(group) > 0:
                h = ut.dict_hist(group)
                hitems = list(h.items())
                hvals = [i[1] for i in hitems]
                maxval = max(hvals)
                g = min([k for k, v in hitems if v == maxval])
                assignment[idx] = g
            else:
                assignment[idx] = None

    if verbose:
        logger.info('rectify %d non-trivial groups' % (len(conflict_groups),))
        logger.info('rectify %d trivial groups' % (ntrivial,))

    num_extra = 0

    if len(conflict_groups) > 0:
        grouped_oldnames_ = conflict_groups
        unique_old_names = ut.unique(ut.flatten(grouped_oldnames_))
        num_new_names = len(grouped_oldnames_)
        num_old_names = len(unique_old_names)
        extra_oldnames = []

        # Create padded dummy values.  This accounts for the case where it is
        # impossible to uniquely map to the old db
        num_extra = num_new_names - num_old_names
        if num_extra > 0:
            extra_oldnames = [
                '%s%d' % (extra_prefix, count) for count in range(num_extra)
            ]
        elif num_extra < 0:
            pass
        else:
            extra_oldnames = []
        assignable_names = unique_old_names + extra_oldnames

        total = len(assignable_names)

        # Allocate assignment matrix
        # Start with a large negative value indicating
        # that you must select from your assignments only
        profit_matrix = -np.ones((total, total), dtype=np.int) * (2 * total)
        # Populate assignment profit matrix
        oldname2_idx = ut.make_index_lookup(assignable_names)
        name_freq_list = [ut.dict_hist(names) for names in grouped_oldnames_]
        # Initialize base profit for using a previously used name
        for rowx, name_freq in enumerate(name_freq_list):
            for name, freq in name_freq.items():
                colx = oldname2_idx[name]
                profit_matrix[rowx, colx] = 1
        # Now add in the real profit
        for rowx, name_freq in enumerate(name_freq_list):
            for name, freq in name_freq.items():
                colx = oldname2_idx[name]
                profit_matrix[rowx, colx] += freq
        # Set a small profit for using an extra name
        extra_colxs = ut.take(oldname2_idx, extra_oldnames)
        profit_matrix[:, extra_colxs] = 1

        # Convert to minimization problem
        big_value = (profit_matrix.max()) - (profit_matrix.min())
        cost_matrix = big_value - profit_matrix

        # Don't use munkres, it is pure python and very slow. Use scipy instead
        indexes = list(zip(*scipy.optimize.linear_sum_assignment(cost_matrix)))

        # Map output to be aligned with input
        rx2_cx = dict(indexes)
        assignment_ = [assignable_names[rx2_cx[rx]] for rx in range(num_new_names)]

        # Reintegrate trivial values
        for idx, g in zip(orig_idxs, assignment_):
            assignment[idx] = g

    for idx, val in enumerate(assignment):
        if val is None:
            assignment[idx] = '%s%d' % (extra_prefix, num_extra)
            num_extra += 1
    return assignment
